import os
import sys
import time
import base64
import requests
import subprocess
import tempfile
import logging
import psycopg
import signal
import socket
import unittest
import json
from unittest.mock import Mock, patch, MagicMock
from fastapi.testclient import TestClient
from fastapi import FastAPI

logging.basicConfig(
    format='[%(levelname)s] %(asctime)s - %(message)s',
    level=logging.INFO
)
logger = logging.getLogger(__name__)

SERVER_HOST = "localhost"
SERVER_PORT = 8083
SERVER_URL = f"http://{SERVER_HOST}:{SERVER_PORT}"

DB_CONN_STRING = "dbname=noetl user=noetl password=noetl host=localhost port=5434"

def check_port(port, host='localhost'):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind((host, port))
        return True
    except socket.error:
        return False
    finally:
        sock.close()

def kill_process(port):
    logger.info(f"Checking for processes using port {port}...")

    if check_port(port):
        logger.info(f"Port {port} is already available")
        return True

    logger.info(f"Port {port} is in use. Attempting to kill the process...")

    try:
        if sys.platform.startswith('darwin') or sys.platform.startswith('linux'):
            cmd = f"lsof -i :{port} -t"
            try:
                output = subprocess.check_output(cmd, shell=True).decode().strip()

                if output:
                    pids = output.split('\n')
                    killed = False

                    for pid in pids:
                        if pid.strip():
                            logger.info(f"Killing process {pid} using port {port}")
                            try:
                                os.kill(int(pid), signal.SIGTERM)
                                time.sleep(1)
                                try:
                                    os.kill(int(pid), 0)
                                    logger.info(f"Process {pid} did not terminate, sending SIGKILL")
                                    os.kill(int(pid), signal.SIGKILL)
                                except OSError:
                                    logger.info(f"Process {pid} terminated successfully")

                                killed = True
                            except OSError as e:
                                logger.warning(f"Error killing process {pid}: {e}")

                    if check_port(port):
                        logger.info(f"Port {port} is now available")
                        return True
                    else:
                        logger.warning(f"Port {port} is still in use after killing processes")
                        return False
            except subprocess.CalledProcessError as e:
                logger.warning(f"Error running lsof command: {e}.")
                time.sleep(2)
                if check_port(port):
                    logger.info(f"Port {port} is now available")
                    return True
        elif sys.platform.startswith('win'):
            cmd = f"netstat -ano | findstr :{port}"
            output = subprocess.check_output(cmd, shell=True).decode()

            if output:
                lines = output.strip().split('\n')
                killed = False

                for line in lines:
                    if f":{port}" in line and "LISTENING" in line:
                        pid = line.strip().split()[-1]
                        logger.info(f"Killing process {pid} using port {port}")
                        try:
                            result = subprocess.call(f"taskkill /F /PID {pid}", shell=True)
                            if result == 0:
                                logger.info(f"Process {pid} terminated successfully")
                                killed = True
                            else:
                                logger.warning(f"Failed to kill process {pid}")
                        except Exception as e:
                            logger.warning(f"Error killing process {pid}: {e}")

                if check_port(port):
                    logger.info(f"Port {port} is now available")
                    return True
                else:
                    logger.warning(f"Port {port} is still in use after killing processes")
                    return False

        logger.warning(f"Could not kill process using port {port}")
        if check_port(port):
            logger.info(f"Port {port} is now available")
            return True
        return False
    except Exception as e:
        logger.error(f"Error killing process on port {port}: {e}")
        if check_port(port):
            logger.info(f"Port {port} is now available despite errors")
            return True
        return False

def start_server():
    logger.info("Starting NoETL server...")
    if not kill_process(SERVER_PORT):
        logger.error(f"Failed to free up port {SERVER_PORT}. Cannot start server.")
        return None, None

    server_log = tempfile.NamedTemporaryFile(delete=False, suffix=".log")
    server_log_path = server_log.name
    server_log.close()

    cmd = [
        "noetl",
        "server",
        "--port", str(SERVER_PORT),
        "--force"
    ]

    logger.info(f"Running command: {' '.join(cmd)}")

    with open(server_log_path, "w") as log_file:
        server_process = subprocess.Popen(
            cmd,
            stdout=log_file,
            stderr=subprocess.STDOUT,
            preexec_fn=os.setsid
        )

    logger.info("Waiting for server to start...")
    max_retries = 10
    retry_interval = 1

    for i in range(max_retries):
        try:
            response = requests.get(f"{SERVER_URL}/health")
            if response.status_code == 200:
                logger.info("Server started successfully")
                return server_process, server_log_path
        except requests.exceptions.ConnectionError:
            pass

        logger.info(f"Waiting for server to start (attempt {i+1}/{max_retries})...")
        time.sleep(retry_interval)
    logger.error("Failed to start server")

    try:
        with open(server_log_path, "r") as log_file:
            logger.error(f"Server log:\n{log_file.read()}")
    except Exception as e:
        logger.error(f"Error reading server log: {e}")

    try:
        os.killpg(os.getpgid(server_process.pid), signal.SIGTERM)
    except Exception as e:
        logger.error(f"Error killing server process: {e}")

    return None, server_log_path

def stop_server(server_process, server_log_path):
    if server_process:
        logger.info("Stopping NoETL server")
        try:
            os.killpg(os.getpgid(server_process.pid), signal.SIGTERM)
            server_process.wait(timeout=5)
            logger.info("Server stopped")
        except Exception as e:
            logger.error(f"Error stopping server: {e}")
            try:
                os.killpg(os.getpgid(server_process.pid), signal.SIGKILL)
            except:
                pass
    if server_log_path and os.path.exists(server_log_path):
        try:
            with open(server_log_path, "r") as log_file:
                logger.info(f"Server log:\n{log_file.read()}")
            os.unlink(server_log_path)
        except Exception as e:
            logger.error(f"Error reading/removing server log: {e}")

def clean_database():
    try:
        conn = psycopg.connect(DB_CONN_STRING)
        with conn.cursor() as cursor:
            cursor.execute(
                "DELETE FROM catalog WHERE resource_path LIKE 'test_%'"
            )
        conn.commit()
        conn.close()
        logger.info("Database cleaned up")
    except Exception as e:
        logger.error(f"Error cleaning database: {e}")

def test_upload_playbook():
    logger.info("Testing playbooks upload")
    test_playbook = """
apiVersion: 0.1.0
kind: Playbook
name: test_playbook
path: test_playbook
spec:
  steps:
    - name: echo
      type: echo
      input:
        message: "Hello, World!"
    """

    playbook_base64 = base64.b64encode(test_playbook.encode()).decode()

    response = requests.post(
        f"{SERVER_URL}/catalog/register",
        json={"content_base64": playbook_base64}
    )

    if response.status_code == 200:
        logger.info("Playbook uploaded")
        logger.info(f"Response: {response.json()}")
        return True
    else:
        logger.error(f"Failed to upload playbooks: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

def test_list_playbooks():
    logger.info("Testing playbooks listing")
    response = requests.get(f"{SERVER_URL}/catalog/list")

    if response.status_code == 200:
        response_data = response.json()
        playbooks = response_data.get("entries", [])
        logger.info(f"Found {len(playbooks)} playbooks")

        test_playbook_found = False
        for playbook in playbooks:
            if playbook.get("resource_path") == "test_playbook":
                test_playbook_found = True
                logger.info("Test playbooks found in catalog")
                break

        if not test_playbook_found:
            logger.error("Test playbooks not found in catalog")
            logger.info(f"Playbooks in catalog: {playbooks}")
            return False

        return True
    else:
        logger.error(f"Failed to list playbooks: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

def test_execute_playbook():
    logger.info("Testing playbooks execution")
    response = requests.get(f"{SERVER_URL}/catalog/list")
    if response.status_code != 200:
        logger.error(f"Failed to list playbooks: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

    playbooks = response.json().get("entries", [])
    test_playbook_version = None
    for playbook in playbooks:
        if playbook.get("resource_path") == "test_playbook":
            test_playbook_version = playbook.get("resource_version")
            break

    if not test_playbook_version:
        logger.error("Test playbooks not found in catalog")
        return False

    logger.info(f"Using test playbooks version: {test_playbook_version}")

    response = requests.post(
        f"{SERVER_URL}/agent/execute",
        json={
            "path": "test_playbook",
            "version": test_playbook_version,
            "input_payload": {"additional_message": "Testing execution"}
        }
    )

    if response.status_code == 200:
        result = response.json()
        logger.info("Playbook executed successfully")
        logger.info(f"Response: {result}")
        if result.get("status") == "success":
            return True
        else:
            logger.error(f"Execution failed: {result.get('error')}")
            return False
    else:
        logger.error(f"Failed to execute playbooks: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

def test_execute_playbook_async():
    logger.info("Testing asynchronous playbooks execution")
    response = requests.get(f"{SERVER_URL}/catalog/list")
    if response.status_code != 200:
        logger.error(f"Failed to list playbooks: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

    playbooks = response.json().get("entries", [])
    test_playbook_version = None
    for playbook in playbooks:
        if playbook.get("resource_path") == "test_playbook":
            test_playbook_version = playbook.get("resource_version")
            break

    if not test_playbook_version:
        logger.error("Test playbooks not found in catalog")
        return False

    logger.info(f"Using test playbooks version: {test_playbook_version}")
    response = requests.post(
        f"{SERVER_URL}/agent/execute-async",
        json={
            "path": "test_playbook",
            "version": test_playbook_version,
            "input_payload": {"additional_message": "Testing async execution"}
        }
    )

    if response.status_code == 200:
        result = response.json()
        logger.info("Playbook execution started asynchronously")
        logger.info(f"Response: {result}")

        event_id = result.get("event_id")
        if not event_id:
            logger.error("No event ID returned")
            return False

        max_retries = 10
        retry_interval = 1

        for i in range(max_retries):
            response = requests.get(f"{SERVER_URL}/events/{event_id}")

            if response.status_code == 200:
                event = response.json()
                state = event.get("state")

                if state in ["COMPLETED", "FAILED", "ERROR"]:
                    logger.info(f"Execution completed with state: {state}")
                    return state == "COMPLETED"

            logger.info(f"Waiting for execution to complete (attempt {i+1}/{max_retries})...")
            time.sleep(retry_interval)

        logger.error("Execution did not complete in time")
        return False
    else:
        logger.error(f"Failed to start async execution: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

def test_get_event_by_query_param():
    logger.info("Testing event retrieval using event_id path endpoint...")
    response = requests.get(f"{SERVER_URL}/catalog/list")
    if response.status_code != 200:
        logger.error(f"Failed to list playbooks: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

    playbooks = response.json().get("entries", [])
    test_playbook_version = None
    for playbook in playbooks:
        if playbook.get("resource_path") == "test_playbook":
            test_playbook_version = playbook.get("resource_version")
            break

    if not test_playbook_version:
        logger.error("Test playbooks not found in catalog")
        return False

    logger.info(f"Using test playbooks version: {test_playbook_version}")
    response = requests.post(
        f"{SERVER_URL}/agent/execute-async",
        json={
            "path": "test_playbook",
            "version": test_playbook_version,
            "input_payload": {"additional_message": "Testing event query"}
        }
    )

    if response.status_code == 200:
        result = response.json()
        logger.info("Playbook execution started asynchronously")
        logger.info(f"Response: {result}")
        event_id = result.get("event_id")
        if not event_id:
            logger.error("No event ID returned")
            return False
        time.sleep(1)

        response = requests.get(f"{SERVER_URL}/events/{event_id}")

        if response.status_code == 200:
            event = response.json()
            logger.info("Event retrieved successfully using event_id path endpoint")
            logger.info(f"Response: {event}")
            return True
        else:
            logger.error(f"Failed to retrieve event using event_id path endpoint: {response.status_code}")
            logger.error(f"Response: {response.text}")
            return False
    else:
        logger.error(f"Failed to start async execution: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

def test_get_resource_with_path_segments():
    logger.info("Testing resource retrieval with path segments")
    test_playbook = """
apiVersion: 0.1.0
kind: Playbook
name: test_playbook_segments
path: test/segments/playbooks
spec:
  steps:
    - name: echo
      type: echo
      input:
        message: "Hello, World!"
    """

    playbook_base64 = base64.b64encode(test_playbook.encode()).decode()

    response = requests.post(
        f"{SERVER_URL}/catalog/register",
        json={"content_base64": playbook_base64}
    )

    if response.status_code == 200:
        result = response.json()
        logger.info("Playbook with path segments uploaded successfully")
        logger.info(f"Response: {result}")

        path = result.get("resource_path")
        version = result.get("resource_version")

        if not path or not version:
            logger.error("No path or version returned")
            return False

        response = requests.get(f"{SERVER_URL}/catalog/{path}/{version}")

        if response.status_code == 200:
            resource = response.json()
            logger.info("Resource retrieved successfully with path segments")
            logger.info(f"Response: {resource}")
            return True
        else:
            logger.error(f"Failed to retrieve resource with path segments: {response.status_code}")
            logger.error(f"Response: {response.text}")
            return False
    else:
        logger.error(f"Failed to upload playbooks with path segments: {response.status_code}")
        logger.error(f"Response: {response.text}")
        return False

# Integration tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# Mock the dependencies before importing
with patch('noetl.server.get_pgdb_connection') as mock_pgdb:
    mock_pgdb.return_value = "mock://connection"
    from noetl.server import CatalogService, EventService

class TestServerAPIIntegration(unittest.TestCase):
    def setUp(self):
        # Mock database connections
        self.psycopg_connect_patcher = patch('noetl.server.psycopg.connect')
        self.mock_connect = self.psycopg_connect_patcher.start()

        self.conn_mock = Mock()
        self.cursor_mock = Mock()
        self.conn_mock.cursor.return_value = MagicMock()
        self.conn_mock.cursor.return_value.__enter__.return_value = self.cursor_mock
        self.mock_connect.return_value = self.conn_mock

        self.addCleanup(self.psycopg_connect_patcher.stop)

        # Initialize services
        self.catalog_service = CatalogService()
        self.event_service = EventService()

    def test_catalog_service_integration(self):
        """Test CatalogService with realistic scenarios"""
        # Test successful resource registration - fix mock setup
        self.cursor_mock.fetchone.side_effect = [(0,), (0,)]  # COUNT queries return integers
        self.cursor_mock.fetchall.return_value = []  # No existing versions

        test_playbook = """
apiVersion: 0.1.0
kind: Playbook
name: test_integration_playbook
path: test_integration_playbook
description: Integration test playbooks
steps:
  - name: log_step
    action: log
    input:
      message: "Integration test message"
"""

        result = self.catalog_service.register_resource(test_playbook)

        # Verify the result structure
        self.assertIsInstance(result, dict)
        self.assertIn('status', result)
        self.assertIn('resource_path', result)
        self.assertIn('resource_version', result)
        self.assertEqual(result['status'], 'success')
        self.assertEqual(result['resource_path'], 'test_integration_playbook')

    def test_catalog_list_entries(self):
        """Test listing catalog entries"""
        mock_entries = [
            ('playbook1', 'playbooks', '0.1.0', '{}', '2023-01-01'),
            ('playbook2', 'playbooks', '0.2.0', '{}', '2023-01-02'),
        ]
        self.cursor_mock.fetchall.return_value = mock_entries

        entries = self.catalog_service.list_entries()

        self.assertIsInstance(entries, list)
        self.assertEqual(len(entries), 2)
        self.assertEqual(entries[0]['resource_path'], 'playbook1')
        self.assertEqual(entries[1]['resource_path'], 'playbook2')

    def test_catalog_list_entries_filtered(self):
        """Test listing catalog entries with resource type filter"""
        mock_entries = [
            ('playbook1', 'playbooks', '0.1.0', '{}', '2023-01-01'),
        ]
        self.cursor_mock.fetchall.return_value = mock_entries

        entries = self.catalog_service.list_entries(resource_type='playbooks')

        self.assertIsInstance(entries, list)
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0]['resource_type'], 'playbooks')

    def test_event_service_emit(self):
        """Test EventService emit functionality"""
        self.cursor_mock.fetchone.return_value = (0,)  # Return count as tuple

        event_data = {
            "event_type": "TEST_EVENT",
            "status": "CREATED",
            "message": "Test event message"
        }

        result = self.event_service.emit(event_data)

        # Verify event_id was added
        self.assertIn('event_id', result)
        self.assertEqual(result['event_type'], 'TEST_EVENT')
        self.assertEqual(result['status'], 'CREATED')

    def test_catalog_version_management(self):
        """Test version increment logic"""
        test_cases = [
            ('0.1.0', '0.1.1'),
            ('1.2.3', '1.2.4'),
            ('0.0.1', '0.0.2'),
            ('10.20.30', '10.20.31')
        ]

        for input_version, expected_output in test_cases:
            with self.subTest(input_version=input_version):
                result = self.catalog_service.increment_version(input_version)
                self.assertEqual(result, expected_output)

    def test_catalog_fetch_with_fallback(self):
        """Test fetch entry with path fallback logic"""
        # First call returns None (path not found), second call finds filename
        self.cursor_mock.fetchone.side_effect = [
            None,  # First query with full path fails
            ('filename.yaml', 'playbooks', '0.1.0', 'content', '{}', '{}')  # Second query with filename succeeds
        ]

        result = self.catalog_service.fetch_entry('some/path/to/filename.yaml', '0.1.0')

        self.assertIsNotNone(result)
        self.assertEqual(result['resource_path'], 'filename.yaml')
        self.assertEqual(result['resource_version'], '0.1.0')

    def test_catalog_error_handling(self):
        """Test error handling in catalog operations"""
        # Test database connection error
        self.mock_connect.side_effect = Exception("Database connection failed")

        version = self.catalog_service.get_latest_version('test_resource')
        self.assertEqual(version, '0.1.0')  # Should return default on error

        # Reset the mock
        self.mock_connect.side_effect = None
        self.mock_connect.return_value = self.conn_mock

    def test_catalog_duplicate_version_handling(self):
        """Test handling of duplicate versions during registration"""
        # Simulate version collision scenario
        self.cursor_mock.fetchone.side_effect = [
            (1,),     # Count query - has existing entries
            ('0.1.0',), # Latest version query
            (1,),     # Check if incremented version exists (collision)
            (0,),     # Check if next incremented version exists (success)
        ]
        self.cursor_mock.fetchall.return_value = [('0.1.0',)]

        test_playbook = """
apiVersion: 0.1.0
kind: Playbook
name: collision_test
path: collision_test
steps:
  - name: test_step
    action: log
"""

        result = self.catalog_service.register_resource(test_playbook)

        # Should succeed with incremented version
        self.assertEqual(result['status'], 'success')
        self.assertEqual(result['resource_path'], 'collision_test')
        # Version should be incremented beyond the collision
        self.assertIn('resource_version', result)

    def test_event_service_error_handling(self):
        """Test error handling in event service"""
        # Simulate database error
        self.mock_connect.side_effect = Exception("Database error")

        event_data = {"event_type": "TEST", "message": "test"}

        with self.assertRaises(Exception):
            self.event_service.emit(event_data)

        # Reset the mock
        self.mock_connect.side_effect = None
        self.mock_connect.return_value = self.conn_mock


class TestServerIntegrationScenarios(unittest.TestCase):
    """Integration tests for complex scenarios"""

    def setUp(self):
        self.psycopg_connect_patcher = patch('noetl.server.psycopg.connect')
        self.mock_connect = self.psycopg_connect_patcher.start()

        self.conn_mock = Mock()
        self.cursor_mock = Mock()
        self.conn_mock.cursor.return_value = MagicMock()
        self.conn_mock.cursor.return_value.__enter__.return_value = self.cursor_mock
        self.mock_connect.return_value = self.conn_mock

        self.addCleanup(self.psycopg_connect_patcher.stop)
        self.catalog_service = CatalogService()

    def test_full_workflow_scenario(self):
        """Test a complete workflow: register -> fetch -> list"""
        # Setup mocks for registration
        self.cursor_mock.fetchone.side_effect = [(0,), (0,)]
        self.cursor_mock.fetchall.return_value = []

        # 1. Register a resource
        playbook_content = """
apiVersion: 0.1.0
kind: Playbook
name: workflow_test
path: workflow_test
description: Full workflow test
steps:
  - name: init
    action: log
    input:
      message: "Workflow started"
"""

        register_result = self.catalog_service.register_resource(playbook_content)
        self.assertEqual(register_result['status'], 'success')

        # 2. Fetch the registered resource - Create new cursor mock for fetch
        fetch_cursor_mock = Mock()
        mock_fetch_result = (
            'workflow_test', 'playbooks', '0.1.1',
            playbook_content, '{}', '{}'
        )
        fetch_cursor_mock.fetchone.return_value = mock_fetch_result

        # Replace the cursor mock temporarily for fetch operation
        with patch.object(self.conn_mock, 'cursor') as mock_cursor_method:
            mock_cursor_method.return_value.__enter__.return_value = fetch_cursor_mock

            fetch_result = self.catalog_service.fetch_entry('workflow_test', '0.1.1')
            self.assertIsNotNone(fetch_result)
            self.assertEqual(fetch_result['resource_path'], 'workflow_test')

        # 3. List all entries - Create new cursor mock for list
        list_cursor_mock = Mock()
        mock_list_result = [
            ('workflow_test', 'playbooks', '0.1.1', '{}', '2023-01-01')
        ]
        list_cursor_mock.fetchall.return_value = mock_list_result

        # Replace the cursor mock temporarily for list operation
        with patch.object(self.conn_mock, 'cursor') as mock_cursor_method:
            mock_cursor_method.return_value.__enter__.return_value = list_cursor_mock

            list_result = self.catalog_service.list_entries()
            self.assertEqual(len(list_result), 1)
            self.assertEqual(list_result[0]['resource_path'], 'workflow_test')


def main():
    logger.info("Starting NoETL server API tests")

    clean_database()

    server_process, server_log_path = start_server()
    if not server_process:
        logger.error("Failed to start server, aborting tests")
        return 1

    try:
        upload_success = test_upload_playbook()
        if not upload_success:
            logger.error("Playbook upload test failed")
            return 1

        list_success = test_list_playbooks()
        if not list_success:
            logger.error("Playbook listing test failed")
            return 1

        execute_success = test_execute_playbook()
        if not execute_success:
            logger.error("Playbook execution test failed")
            return 1

        async_execute_success = test_execute_playbook_async()
        if not async_execute_success:
            logger.error("Asynchronous playbooks execution test failed")
            return 1

        event_query_success = test_get_event_by_query_param()
        if not event_query_success:
            logger.error("Event query parameter test failed")
            return 1

        path_segments_success = test_get_resource_with_path_segments()
        if not path_segments_success:
            logger.error("Resource path segments test failed")
            return 1

        logger.info("All tests passed successfully!")
        return 0

    finally:
        stop_server(server_process, server_log_path)
        clean_database()

if __name__ == "__main__":
    sys.exit(main())



import yaml
import json
import os
import uuid
import datetime
from typing import Dict, List, Any, Optional, Tuple
from jinja2 import Environment, StrictUndefined, BaseLoader
from noetl.render import render_template
from noetl.secret import SecretManager
from noetl.sqlcmd import *
from noetl.common import setup_logger

logger = setup_logger(__name__, include_location=True)

class Worker:

    def __init__(self, playbook_path: str, mock_mode: bool = True, pgdb: str = None):
        logger.debug("=== WORKER.__INIT__: Function entry ===")
        logger.debug(f"WORKER.__INIT__: Parameters - playbook_path={playbook_path}, mock_mode={mock_mode}, pgdb={pgdb}")
        
        self.playbook_path = playbook_path
        self.mock_mode = mock_mode
        self.execution_id = str(uuid.uuid4())
        logger.debug(f"WORKER.__INIT__: Generated execution_id={self.execution_id}")
        
        logger.debug("WORKER.__INIT__: Loading playbook")
        self.playbook = self.load_playbook()
        logger.debug(f"WORKER.__INIT__: Loaded playbook with name={self.playbook.get('name', 'Unnamed')}")
        
        logger.debug("WORKER.__INIT__: Setting up Jinja environment")
        self.jinja_env = Environment(loader=BaseLoader(), undefined=StrictUndefined)
        self.jinja_env.filters['to_json'] = lambda obj: json.dumps(obj)
        self.jinja_env.filters['b64encode'] = lambda s: __import__('base64').b64encode(s.encode('utf-8')).decode('utf-8') if isinstance(s, str) else __import__('base64').b64encode(str(s).encode('utf-8')).decode('utf-8')
        self.jinja_env.globals['now'] = lambda: datetime.datetime.now().isoformat()
        self.jinja_env.globals['env'] = os.environ
        self.next_step_with = None
        
        logger.debug("WORKER.__INIT__: Initializing SecretManager")
        self.secret_manager = SecretManager(self.jinja_env, self.mock_mode)

        if pgdb is None:
            pgdb = f"dbname={os.environ.get('POSTGRES_DB', 'noetl')} user={os.environ.get('POSTGRES_USER', 'noetl')} password={os.environ.get('POSTGRES_PASSWORD', 'noetl')} host={os.environ.get('POSTGRES_HOST', 'localhost')} port={os.environ.get('POSTGRES_PORT', '5434')}"
            logger.info(f"WORKER.__INIT__: Using default Postgres connection: {pgdb}")
        else:
            logger.info(f"WORKER.__INIT__: Using provided Postgres connection: {pgdb}")

        self.pgdb = pgdb
        logger.debug("WORKER.__INIT__: Importing DatabaseSchema")
        from noetl.schema import DatabaseSchema
        
        logger.debug("WORKER.__INIT__: Initializing DatabaseSchema")
        self.db_schema = DatabaseSchema(pgdb=pgdb)
        self.conn = self.db_schema.conn
        
        logger.debug("WORKER.__INIT__: Initializing database")
        self.db_schema.init_database()

        logger.debug("WORKER.__INIT__: Setting up initial context")
        self.context = {
            'job': {
                'uuid': self.execution_id,
                'id': self.execution_id
            }
        }

        logger.debug("WORKER.__INIT__: Loading workload from database")
        db_workload = self.load_workload()
        if db_workload:
            logger.info(f"WORKER.__INIT__: Using workload from database for execution {self.execution_id}")
            logger.debug(f"WORKER.__INIT__: Database workload: {db_workload}")
            rendered_workload = render_template(self.jinja_env, db_workload, self.context)
            logger.debug(f"WORKER.__INIT__: Rendered workload: {rendered_workload}")
            self.context.update(rendered_workload)
        else:
            logger.info(f"WORKER.__INIT__: Workload not found in database. Using default from playbook.")
            playbook_workload = self.playbook.get('workload', {})
            logger.debug(f"WORKER.__INIT__: Playbook workload: {playbook_workload}")
            rendered_workload = render_template(self.jinja_env, playbook_workload, self.context)
            logger.debug(f"WORKER.__INIT__: Rendered workload: {rendered_workload}")
            self.context.update(rendered_workload)

        logger.debug("WORKER.__INIT__: Parsing playbook")
        self.parse_playbook()
        logger.debug("=== WORKER.__INIT__: Function exit ===")
        
        logger.info("=== ENVIRONMENT VARIABLES ===")
        for key, value in sorted(os.environ.items()):
            logger.info(f"ENV: {key}={value}")
        logger.info("=== END ENVIRONMENT VARIABLES ===")

    def store_workload(self, data: Dict):
        """
        Store workload data in the database.

        Args:
            data: The workload
        """
        if not isinstance(data, dict):
            logger.error(f"Invalid workload data type: {type(data)}. Expected dict.")
            return

        try:
            data_json = json.dumps(data)
            logger.info(f"Storing workload data for execution {self.execution_id}: {data_json[:100]}.")
            with self.conn.cursor() as cursor:
                cursor.execute(WORKLOAD_TABLE_EXISTS_POSTGRES)
                table_exists = cursor.fetchone()[0]
                if not table_exists:
                    logger.error("Workload table does not exist in the database.")
                    self.db_schema.init_database()
                    self.conn.commit()
                    logger.info("Database schema initialized.")

            with self.conn.cursor() as cursor:
                logger.info(f"Executing INSERT INTO workload for execution_id: {self.execution_id}")
                cursor.execute(WORKLOAD_INSERT_POSTGRES, (self.execution_id, data_json))
                self.conn.commit()
                logger.info(f"INSERT INTO workload completed and committed for execution_id: {self.execution_id}")
                cursor.execute(WORKLOAD_COUNT_BY_ID_POSTGRES, (self.execution_id,))
                count = cursor.fetchone()[0]
                if count == 0:
                    logger.error(f"Failed to store workload data for execution {self.execution_id}.")
                else:
                    logger.info(f"Stored workload data for execution {self.execution_id}. Found {count} rows.")

                    cursor.execute(WORKLOAD_SELECT_POSTGRES, (self.execution_id,))
                    row = cursor.fetchone()
                    if row and row[0]:
                        logger.info(f"Retrieved workload data for execution {self.execution_id}: {row[0][:100]}.")
                    else:
                        logger.error(f"Failed to retrieve workload data for execution {self.execution_id} count: {count}.")
        except Exception as e:
            logger.error(f"Error storing workload data: {e}", exc_info=True)

    def load_workload(self) -> Dict:
        """
        Load workload data from the database.

        Returns:
            The workload data
        """
        try:
            logger.info(f"Loading workload data for execution {self.execution_id}")
            with self.conn.cursor() as cursor:
                logger.info("Checking if workload table exists")
                cursor.execute(WORKLOAD_TABLE_EXISTS_POSTGRES)
                table_exists = cursor.fetchone()[0]
                if not table_exists:
                    logger.error("Workload table does not exist in the database")
                    return {}
                else:
                    logger.info("Workload table exists in the database.")
                logger.info("Checking if workload table has rows")
                cursor.execute(WORKLOAD_COUNT_POSTGRES)
                total_count = cursor.fetchone()[0]
                logger.info(f"Total rows in workload table: {total_count}.")
                logger.info(f"Querying workload data for execution_id: {self.execution_id}")
                cursor.execute(WORKLOAD_SELECT_POSTGRES, (self.execution_id,))
                row = cursor.fetchone()

                if row and row[0]:
                    logger.info(f"Found workload data for execution {self.execution_id}")
                    try:
                        workload_data = json.loads(row[0])
                        logger.info(f"Parsed workload data for execution {self.execution_id}: {str(workload_data)[:100]}.")
                        return workload_data
                    except json.JSONDecodeError as e:
                        logger.error(f"Error decoding workload data: {e}.")
                        logger.error(f"Raw data: {row[0][:100]}.")
                        return {}
                else:
                    logger.info(f"No workload data found for execution {self.execution_id}")
                    cursor.execute(WORKLOAD_SELECT_ALL_IDS_POSTGRES)
                    all_execution_ids = [r[0] for r in cursor.fetchall()]
                    logger.info(f"All execution_ids in workload table: {all_execution_ids}")
                    return {}
        except Exception as e:
            logger.error(f"Error loading workload data: {e}", exc_info=True)
            return {}

    def load_playbook(self) -> Dict:
        """
        Load the playbooks from the file.

        Returns:
            The playbooks data
        """
        if not os.path.exists(self.playbook_path):
            raise FileNotFoundError(f"Playbook not found: {self.playbook_path}")

        with open(self.playbook_path, 'r') as f:
            return yaml.safe_load(f)

    def parse_playbook(self):
        """
        Parse the playbooks and store in the database.
        """
        logger.debug("=== WORKER.PARSE_PLAYBOOK: Function entry ===")
        logger.debug(f"WORKER.PARSE_PLAYBOOK: Parsing playbook for execution_id={self.execution_id}")
        
        with self.conn.cursor() as cursor:
            workflow_steps = self.playbook.get('workflow', [])
            logger.debug(f"WORKER.PARSE_PLAYBOOK: Found {len(workflow_steps)} workflow steps")
            
            for step in workflow_steps:
                step_id = str(uuid.uuid4())
                step_name = step.get('step')
                step_type = 'standard'
                if 'loop' in step:
                    step_type = 'loop'
                elif 'end_loop' in step:
                    step_type = 'end_loop'
                elif 'call' in step:
                    step_type = 'call'
                
                logger.debug(f"WORKER.PARSE_PLAYBOOK: Processing step={step_name}, type={step_type}")

                params = (
                    self.execution_id,
                    step_id,
                    step_name,
                    step_type,
                    step.get('desc', ''),
                    json.dumps(step)
                )
                logger.debug(f"WORKER.PARSE_PLAYBOOK: Inserting step into workflow table with params={params}")
                cursor.execute(WORKFLOW_INSERT_POSTGRES, params)

                next_steps = step.get('next', [])
                if not isinstance(next_steps, list):
                    next_steps = [next_steps]
                
                logger.debug(f"WORKER.PARSE_PLAYBOOK: Step {step_name} has {len(next_steps)} next steps")

                for next_step in next_steps:
                    logger.debug(f"WORKER.PARSE_PLAYBOOK: Processing next_step={next_step} for step={step_name}")
                    
                    if isinstance(next_step, dict):
                        if 'when' in next_step and 'then' in next_step:
                            condition = next_step.get('when')
                            then_steps = next_step.get('then', [])
                            if not isinstance(then_steps, list):
                                then_steps = [then_steps]
                            
                            logger.debug(f"WORKER.PARSE_PLAYBOOK: Found conditional transition with condition={condition} and {len(then_steps)} then steps")

                            for then_step in then_steps:
                                if isinstance(then_step, dict):
                                    to_step = then_step.get('step')
                                    with_params = json.dumps(then_step.get('with', {}))
                                    logger.debug(f"WORKER.PARSE_PLAYBOOK: Then step is a dict with step={to_step}, with_params={with_params}")
                                else:
                                    to_step = then_step
                                    with_params = '{}'
                                    logger.debug(f"WORKER.PARSE_PLAYBOOK: Then step is a string: {to_step}")

                                if to_step is not None and to_step != '':
                                    params = (
                                        self.execution_id,
                                        step_name,
                                        to_step,
                                        condition,
                                        with_params
                                    )
                                    logger.debug(f"WORKER.PARSE_PLAYBOOK: Inserting conditional transition with params={params}")
                                    cursor.execute(TRANSITION_INSERT_CONDITION_POSTGRES, params)
                                else:
                                    logger.warning(f"WORKER.PARSE_PLAYBOOK: Skipping transition with null to_step from {step_name} with condition {condition}")
                        elif 'else' in next_step:
                            else_steps = next_step.get('else', [])
                            if not isinstance(else_steps, list):
                                else_steps = [else_steps]
                            
                            logger.debug(f"WORKER.PARSE_PLAYBOOK: Found else transition with {len(else_steps)} else steps")

                            for else_step in else_steps:
                                if isinstance(else_step, dict):
                                    to_step = else_step.get('step')
                                    with_params = json.dumps(else_step.get('with', {}))
                                    logger.debug(f"WORKER.PARSE_PLAYBOOK: Else step is a dict with step={to_step}, with_params={with_params}")
                                else:
                                    to_step = else_step
                                    with_params = '{}'
                                    logger.debug(f"WORKER.PARSE_PLAYBOOK: Else step is a string: {to_step}")

                                if to_step is not None and to_step != '':
                                    params = (
                                        self.execution_id,
                                        step_name,
                                        to_step,
                                        'else',
                                        with_params
                                    )
                                    logger.debug(f"WORKER.PARSE_PLAYBOOK: Inserting else transition with params={params}")
                                    cursor.execute(TRANSITION_INSERT_CONDITION_POSTGRES, params)
                                else:
                                    logger.warning(f"WORKER.PARSE_PLAYBOOK: Skipping transition with null to_step from {step_name} with condition 'else'")
                    elif isinstance(next_step, str):
                        logger.debug(f"WORKER.PARSE_PLAYBOOK: Next step is a string: {next_step}")
                        if next_step is not None and next_step != '':
                            params = (
                                self.execution_id,
                                step_name,
                                next_step,
                                '',
                                '{}'
                            )
                            logger.debug(f"WORKER.PARSE_PLAYBOOK: Inserting simple transition with params={params}")
                            cursor.execute(TRANSITION_INSERT_CONDITION_POSTGRES, params)
                        else:
                            logger.warning(f"WORKER.PARSE_PLAYBOOK: Skipping transition with null next_step from {step_name}")
                    else:
                        to_step = next_step.get('step') if next_step else None
                        with_params = json.dumps(next_step.get('with', {})) if next_step else '{}'
                        logger.debug(f"WORKER.PARSE_PLAYBOOK: Next step is an object with step={to_step}, with_params={with_params}")
                        
                        if to_step is not None and to_step != '':
                            params = (
                                self.execution_id,
                                step_name,
                                to_step,
                                '',
                                with_params
                            )
                            logger.debug(f"WORKER.PARSE_PLAYBOOK: Inserting object transition with params={params}")
                            cursor.execute(TRANSITION_INSERT_CONDITION_POSTGRES, params)
                        else:
                            logger.warning(f"WORKER.PARSE_PLAYBOOK: Skipping transition with null to_step from {step_name}")

            workbook_tasks = self.playbook.get('workbook', [])
            logger.debug(f"WORKER.PARSE_PLAYBOOK: Found {len(workbook_tasks)} workbook tasks")
            
            for task in workbook_tasks:
                task_id = str(uuid.uuid4())
                task_name = task.get('name') or task.get('task')
                task_type = task.get('type', 'http')

                params = (
                    self.execution_id,
                    task_id,
                    task_name,
                    task_type,
                    json.dumps(task)
                )
                cursor.execute(WORKBOOK_INSERT_POSTGRES, params)

            self.conn.commit()

    def update_context(self, key: str, value: Any):
        """
        Update the context with a new kv.

        Args:
            key: key
            value: value to set for the key
        """
        self.context[key] = value
        self.log_event(
            'context_update', f"{self.execution_id}_context_{key}", key, 'context',
            'success', 0, self.get_context(), None,
            {'context_updated': True}, None,
            worker_id=os.environ.get('WORKER_ID', 'local'),
            distributed_state='updated',
            context_key=key,
            context_value=value
        )

    def get_context(self, include_locals: Dict = None) -> Dict:
        """
        Get the current context.

        Args:
            include_locals: local variables to add to the context

        Returns:
            The current context
        """
        if include_locals:
            return {**self.context, **include_locals}
        return self.context

    def log_event(self, event_type: str, node_id: str, node_name: str, node_type: str,
                  status: str, duration: float, input_context: Dict, output_result: Any,
                  metadata: Dict = None, parent_event_id: str = None,
                  loop_id: str = None, loop_name: str = None, iterator: str = None,
                  items: List[Any] = None, current_index: int = None, current_item: Any = None,
                  results: List[Any] = None, worker_id: str = None, distributed_state: str = None,
                  error: str = None, context_key: str = None, context_value: Any = None):
        """
        Log an event to the database.

        Args:
            event_type: The type of event
            node_id: The ID of the node
            node_name: The name of the node
            node_type: The type of node
            status: The status of the event
            duration: The duration of the event
            input_context: The input context
            output_result: The output result
            metadata: Additional metadata
            parent_event_id: The ID of the parent event
            loop_id: The ID of the loop
            loop_name: The name of the loop
            iterator: The iterator variable name
            items: The items being iterated over
            current_index: The current index in the loop
            current_item: The current item in the loop
            results: The results of the loop
            worker_id: The ID of the worker
            distributed_state: The distributed state
            error: The error message
            context_key: The context key being updated
            context_value: The context value being set

        Returns:
            The ID of the event
        """
        event_id = str(uuid.uuid4())

        input_context_serial = json.dumps({k: v for k, v in input_context.items() if not k.startswith('_')})
        output_result_serial = json.dumps(output_result) if output_result is not None else None
        metadata_serial = json.dumps(metadata) if metadata is not None else None
        items_serial = json.dumps(items) if items is not None else None
        current_item_serial = json.dumps(current_item) if current_item is not None else None
        results_serial = json.dumps(results) if results is not None else None
        context_value_serial = json.dumps(context_value) if context_value is not None and not isinstance(context_value, (str, int, float, bool)) else str(context_value) if context_value is not None else None

        params = (
            self.execution_id, event_id, parent_event_id, datetime.datetime.now(), event_type,
            node_id, node_name, node_type, status, duration,
            input_context_serial, output_result_serial, metadata_serial, error,
            loop_id, loop_name, iterator, items_serial, current_index, current_item_serial, results_serial,
            worker_id, distributed_state, context_key, context_value_serial
        )

        with self.conn.cursor() as cursor:
            cursor.execute(EVENT_LOG_INSERT_POSTGRES, params)
            self.conn.commit()

        return event_id

    def save_step_result(self, step_id: str, step_name: str, parent_id: str,
                         status: str, data: Any = None, error: str = None):
        """
        Save the result of a step.

        Args:
            step_id: The ID of the step
            step_name: The name of the step
            parent_id: The ID of the parent step
            status: The status of the step
            data: The data returned by the step
            error: The error message

        Returns:
            The ID of the step
        """
        event_id = self.log_event(
            'step_result', step_id, step_name, 'step',
            status, 0, self.get_context(), data,
            {'parent_id': parent_id}, parent_id,
            worker_id=os.environ.get('WORKER_ID', 'local'),
            distributed_state=status,
            error=error
        )

        return step_id

    def create_loop_state(self, loop_name: str, iterator: str, items: List[Any], parent_id: str = None):
        """
        Create a new loop state.

        Args:
            loop_name: The name of the loop
            iterator: The iterator variable name
            items: The items to iterate over
            parent_id: The ID of the parent step

        Returns:
            The ID of the loop
        """
        loop_id = str(uuid.uuid4())

        self.log_event(
            'loop_state_created', loop_id, loop_name, 'step.loop',
            'created', 0, self.get_context(), None,
            {'state_created': True}, parent_id,
            loop_id=loop_id, loop_name=loop_name, iterator=iterator,
            items=items, current_index=-1, current_item=None, results=[],
            worker_id=os.environ.get('WORKER_ID', 'local'),
            distributed_state='created'
        )

        return loop_id

    def update_loop_state(self, loop_id: str, current_index: int, current_item: Any, status: str):
        """
        Update the state of a loop.

        Args:
            loop_id: The ID of the loop
            current_index: The current index in the loop
            current_item: The current item in the loop
            status: The status of the loop
        """
        params = (self.execution_id, loop_id)
        with self.conn.cursor() as cursor:
            cursor.execute(LOOP_SELECT_POSTGRES, params)
            row = cursor.fetchone()

        if row:
            loop_name, iterator, items, results = row

            self.log_event(
                'loop_state_updated', loop_id, loop_name, 'step.loop',
                'success', 0, self.get_context(), None,
                {'state_updated': True}, None,
                loop_id=loop_id, loop_name=loop_name, iterator=iterator,
                items=json.loads(items) if items else None,
                current_index=current_index,
                current_item=current_item,
                results=json.loads(results) if results else [],
                worker_id=os.environ.get('WORKER_ID', 'local'),
                distributed_state=status
            )

    def add_loop_result(self, loop_id: str, result: Any):
        """
        Add a result to a loop.

        Args:
            loop_id: The ID of the loop
            result: The result to add
        """
        params = (self.execution_id, loop_id)
        with self.conn.cursor() as cursor:
            cursor.execute(LOOP_DETAILS_SELECT_POSTGRES, params)
            row = cursor.fetchone()

        if row:
            loop_name, iterator, items, current_index, current_item, results = row

            current_results = json.loads(results) if results else []
            current_results.append(result)
            self.log_event(
                'loop_result_added', loop_id, loop_name, 'step.loop',
                'success', 0, self.get_context(), result,
                {'result_added': True}, None,
                loop_id=loop_id, loop_name=loop_name, iterator=iterator,
                items=json.loads(items) if items else None,
                current_index=current_index,
                current_item=json.loads(current_item) if current_item else None,
                results=current_results,
                worker_id=os.environ.get('WORKER_ID', 'local'),
                distributed_state='result_added'
            )

    def get_loop_results(self, loop_id: str) -> List[Any]:
        """
        Get the results of a loop.

        Args:
            loop_id: The ID of the loop

        Returns:
            The results of the loop
        """
        params = (self.execution_id, loop_id)
        with self.conn.cursor() as cursor:
            cursor.execute(LOOP_RESULTS_SELECT_POSTGRES, params)
            result_row = cursor.fetchone()

        if result_row and result_row[0]:
            return json.loads(result_row[0])
        return []

    def get_active_loops(self) -> List[Dict]:
        """
        Get all active loops.

        Returns:
            A list of active loops
        """
        params = (self.execution_id, self.execution_id)
        with self.conn.cursor() as cursor:
            cursor.execute(GET_ACTIVE_LOOPS_POSTGRES, params)
            rows = cursor.fetchall()

        loops = []
        for row in rows:
            loops.append({
                'id': row[0],
                'name': row[1],
                'iterator': row[2],
                'items': json.loads(row[3]) if row[3] else [],
                'index': row[4],
                'current_item': json.loads(row[5]) if row[5] else None,
                'results': json.loads(row[6]) if row[6] else []
            })

        return loops

    def find_loop(self, loop_name: str, include_completed: bool = False) -> Optional[Dict]:
        """
        Find a loop by name.

        Args:
            loop_name: The name of the loop
            include_completed: Whether to include completed loops

        Returns:
            The loop information
        """
        logger.debug(f"Loop name: {loop_name}, include_completed: {include_completed}")
        params = (self.execution_id,)
        with self.conn.cursor() as cursor:
            cursor.execute(GET_ALL_LOOPS_POSTGRES, params)
            all_loops = cursor.fetchall()

        logger.debug(f"Loops in database: {all_loops}")
        params = (self.execution_id, loop_name, self.execution_id, loop_name)
        with self.conn.cursor() as cursor:
            if include_completed:
                cursor.execute(FIND_LOOP_INCLUDE_COMPLETED_POSTGRES, params)
            else:
                cursor.execute(FIND_LOOP_EXCLUDE_COMPLETED_POSTGRES, params)
            row = cursor.fetchone()

        if row:
            return {
                'id': row[0],
                'iterator': row[1],
                'items': json.loads(row[2]) if row[2] else [],
                'index': row[3],
                'current_item': json.loads(row[4]) if row[4] else None,
                'results': json.loads(row[5]) if row[5] else []
            }
        return None

    def complete_loop(self, loop_id: str):
        """
        Mark a loop as completed.

        Args:
            loop_id: The ID of the loop
        """
        params = (self.execution_id, loop_id)
        with self.conn.cursor() as cursor:
            cursor.execute(COMPLETE_LOOP_SELECT_POSTGRES, params)
            row = cursor.fetchone()

        if row:
            loop_name, iterator, items, current_index, results = row

            self.log_event(
                'loop_complete', loop_id, loop_name, 'step.loop',
                'success', 0, self.get_context(), None,
                {'completed_by': 'complete_loop'}, None,
                loop_id=loop_id, loop_name=loop_name, iterator=iterator,
                items=json.loads(items) if items else None,
                current_index=current_index,
                current_item=None,
                results=json.loads(results) if results else None,
                worker_id=os.environ.get('WORKER_ID', 'local'),
                distributed_state='completed'
            )

    def store_transition(self, params: Tuple):
        """
        Store a transition in the database.

        Args:
            params: The parameters for the transition
        """
        with self.conn.cursor() as cursor:
            cursor.execute(TRANSITION_INSERT_POSTGRES, params)
            self.conn.commit()

    def get_step_results(self) -> Dict[str, Any]:
        """
        Get the results of all steps.

        Returns:
            A dictionary mapping step names to their results
        """
        step_result = {}
        params = (self.execution_id,)
        with self.conn.cursor() as cursor:
            cursor.execute(STEP_RESULTS_SELECT_POSTGRES, params)
            rows = cursor.fetchall()

        for row in rows:
            if row[1]:
                step_result[row[0]] = json.loads(row[1])

        return step_result

    def export_execution_data(self, filepath: str = None):
        """
        Export execution data to a file.

        Args:
            filepath: The path to the file to export to

        Returns:
            The path to the exported file
        """
        if not filepath:
            filepath = f"noetl_execution_{self.execution_id}.json"

        params = (self.execution_id,)
        with self.conn.cursor() as cursor:
            cursor.execute(EXPORT_EXECUTION_DATA_POSTGRES, params)
            rows = cursor.fetchall()

            if rows:
                columns = [desc[0] for desc in cursor.description]
                event_log_data = [dict(zip(columns, row)) for row in rows]

                with open(filepath, 'w') as f:
                    json.dump(event_log_data, f, default=str, indent=2)

                logger.info(f"Execution data exported to {filepath}")
                return filepath

        logger.warning("No data to export")
        return None

    def find_step(self, step_name: str) -> Optional[Dict]:
        """
        Find a step by name.

        Args:
            step_name: The name of the step

        Returns:
            The step configuration
        """
        for step in self.playbook.get('workflow', []):
            if step.get('step') == step_name:
                return step
        return None

    def find_task(self, task_name: str) -> Optional[Dict]:
        """
        Find a task by name.

        Args:
            task_name: The name of the task

        Returns:
            The task configuration
        """
        for task in self.playbook.get('workbook', []):
            if task.get('name') == task_name or task.get('task') == task_name:
                if 'call' in task:
                    call_config = task['call'].copy()
                    logger.debug(f"Task '{task_name}' has 'call' attribute with type: {call_config.get('type', 'workbook')}")
                    merged_fields = []
                    for key, value in task.items():
                        if key != 'call' and key not in call_config:
                            call_config[key] = value
                            merged_fields.append(key)

                    if merged_fields:
                        logger.debug(f"Merged fields from task into call: {', '.join(merged_fields)}")

                    return call_config
                else:
                    logger.debug(f"Task '{task_name}' using direct attributes with type: {task.get('type', 'http')}")
                    return task
        return None

    def ml_recommendation(self, current_step: str, context: Dict) -> str:
        """
        TODO: future feature. Placeholder for ML recommendation.
        Get a recommendation for the next step using ML.

        Args:
            current_step: The current step
            context: The current context

        Returns:
            The recommended next step
        """
        return None

    def run(self, mlflow: bool = False) -> Dict[str, Any]:
        """
        Run the agent workflow.

        Args:
            mlflow: Flag to use ML flow for workflow control

        Returns:
            A dictionary of workflow results
        """
        from noetl.broker import Broker
        server_url = os.environ.get('NOETL_SERVER_URL', 'http://localhost:8082')
        daemon = Broker(self, server_url=server_url)
        return daemon.run(mlflow)



import re
import json
import logging
import base64
from typing import Any, Dict, List, Union
from jinja2 import Environment, meta, StrictUndefined, BaseLoader

logger = logging.getLogger(__name__)

def add_b64encode_filter(env: Environment) -> Environment:
    """
    Add the b64encode filter to a Jinja2 environment.
    
    Args:
        env: The Jinja2 environment
        
    Returns:
        The Jinja2 environment with the b64encode filter added
    """
    if 'b64encode' not in env.filters:
        env.filters['b64encode'] = lambda s: base64.b64encode(s.encode('utf-8')).decode('utf-8') if isinstance(s, str) else base64.b64encode(str(s).encode('utf-8')).decode('utf-8')
    return env


def render_template(env: Environment, template: Any, context: Dict, rules: Dict = None) -> Any:
    """
    NoETL Jinja2 rendering.

    Args:
        env: The Jinja2 environment
        template: The template to render
        context: The context to use for rendering
        rules: Additional rules for rendering

    Returns:
        The rendered template
    """
    # Ensure the environment has the b64encode filter
    env = add_b64encode_filter(env)
    if isinstance(template, str) and '{{' in template and '}}' in template:
        logger.debug(f"Render template: {template}")
        logger.debug(f"Render template context keys: {list(context.keys())}")
        if 'city' in context:
            logger.debug(f"Render template city value: {context['city']}, Type: {type(context['city'])}")
        if rules:
            logger.debug(f"Render template rules: {rules}")

    render_ctx = dict(context)
    if rules:
        render_ctx.update(rules)

    if isinstance(template, str) and '{{' in template and '}}' in template:
        try:
            expr = template.strip()
            if expr == '{{}}':
                return ""
            if expr.startswith('{{') and expr.endswith('}}'):
                var_path = expr[2:-2].strip()
                if not any(op in var_path for op in ['==', '!=', '<', '>', '+', '-', '*', '/', '|', ' if ', ' else ']):
                    if '.' not in var_path and var_path.strip() in render_ctx:
                        return render_ctx.get(var_path.strip())
                    elif '.' in var_path:
                        parts = var_path.split('.')
                        value = render_ctx
                        valid_path = True
                        for part in parts:
                            part = part.strip()
                            if isinstance(value, dict) and part in value:
                                value = value.get(part)
                            else:
                                valid_path = False
                                break
                        if valid_path:
                            return value

            template_obj = env.from_string(template)
            try:
                rendered = template_obj.render(**render_ctx)
            except Exception as e:
                logger.error(f"Template rendering error: {e}, template: {template}")
                return None

            if (rendered.startswith('[') and rendered.endswith(']')) or \
                    (rendered.startswith('{') and rendered.endswith('}')):
                try:
                    return json.loads(rendered)
                except json.JSONDecodeError:
                    pass

            if rendered.strip() == "":
                return ""

            return rendered
        except Exception as e:
            logger.error(f"Template rendering error: {e}, template: {template}")
            return ""
    elif isinstance(template, dict):
        if not template:
            return template
        return {k: render_template(env, v, render_ctx, rules) for k, v in template.items()}
    elif isinstance(template, list):
        return [render_template(env, item, render_ctx, rules) for item in template]
    return template


class SQLCommentPreserver:
    """Helper class to preserve SQL comments during template rendering."""

    def __init__(self):
        self.comment_placeholders = {}
        self.placeholder_counter = 0

    def preserve_comments(self, sql: str) -> str:
        """Preserve comments with placeholders."""
        self.comment_placeholders.clear()
        self.placeholder_counter = 0

        comment_patterns = [
            r'--[^\n]*',  # Single-line comments
            r'/\*.*?\*/',  # Multi-line comments
        ]

        result = sql
        for pattern in comment_patterns:
            def replace_comment(match):
                comment = match.group(0)
                placeholder = f"__NOETL_COMMENT_PLACEHOLDER_{self.placeholder_counter}__"
                self.comment_placeholders[placeholder] = comment
                self.placeholder_counter += 1
                return placeholder

            result = re.sub(pattern, replace_comment, result, flags=re.DOTALL)

        return result

    def restore_comments(self, sql: str) -> str:
        """Restore SQL comments from placeholders."""
        result = sql
        for placeholder, comment in self.comment_placeholders.items():
            result = result.replace(placeholder, comment)
        return result


def render_sql_template(env: Environment, sql_template: str, context: Dict) -> str:
    """
    Render SQL template while preserving comments and handling multiline strings properly.

    Args:
        env: Jinja2 environment
        sql_template: SQL template string with Jinja2 expressions
        context: Template context

    Returns:
        Rendered SQL string with comments preserved
    """
    # Ensure the environment has the b64encode filter
    env = add_b64encode_filter(env)
    if not sql_template or not isinstance(sql_template, str):
        return sql_template

    if '{{' not in sql_template and '}}' not in sql_template:
        return sql_template

    try:
        comment_preserver = SQLCommentPreserver()
        sql_with_placeholders = comment_preserver.preserve_comments(sql_template)
        template_obj = env.from_string(sql_with_placeholders)
        rendered_sql = template_obj.render(**context)
        final_sql = comment_preserver.restore_comments(rendered_sql)
        logger.debug(f"SQL template rendered, length: {len(final_sql)}")
        return final_sql

    except Exception as e:
        logger.error(f"Error rendering SQL template: {e}")
        logger.debug(f"Failed template: {sql_template[:200]}...")
        raise


def render_duckdb_commands(env: Environment, commands: Union[str, List[str]], context: Dict, task_with: Dict = None) -> List[str]:
    """
    Render DuckDB commands.

    Args:
        env: Jinja2 environment
        commands: SQL commands, string or list
        context: Template context
        task_with: Additional task parameters

    Returns:
        Rendered SQL commands list
    """
    if task_with is None:
        task_with = {}

    full_context = {**context, **task_with}

    if isinstance(commands, str):
        rendered_sql = render_sql_template(env, commands, full_context)

        commands_list = []
        current_command = ""
        in_string = False
        string_char = None

        i = 0
        while i < len(rendered_sql):
            char = rendered_sql[i]

            if char in ('"', "'") and (i == 0 or rendered_sql[i-1] != '\\'):
                if not in_string:
                    in_string = True
                    string_char = char
                elif char == string_char:
                    in_string = False
                    string_char = None

            if char == ';' and not in_string:
                command = current_command.strip()
                if command:
                    commands_list.append(command)
                current_command = ""
            else:
                current_command += char

            i += 1

        if current_command.strip():
            commands_list.append(current_command.strip())

        return [cmd for cmd in commands_list if cmd and not cmd.strip().startswith('--')]

    elif isinstance(commands, list):
        rendered_commands = []
        for cmd in commands:
            if isinstance(cmd, str):
                rendered_cmd = render_sql_template(env, cmd, full_context)
                if rendered_cmd and not rendered_cmd.strip().startswith('--'):
                    rendered_commands.append(rendered_cmd)
            else:
                rendered_commands.append(cmd)
        return rendered_commands

    return []


# def render_template_bool(jinja_env: Environment, template_str: str, context: Dict) -> bool:
#     """
#     Render a template and return as boolean, with proper error handling.
#
#     Args:
#         jinja_env: Jinja2 environment
#         template_str: Template string
#         context: Template context
#
#     Returns:
#         Boolean result or False on error
#     """
#     try:
#         # Check if all referenced variables exist
#         ast = jinja_env.parse(template_str)
#         referenced = meta.find_undeclared_variables(ast)
#         for key in referenced:
#             parts = key.split('.')
#             ctx = context
#             valid_path = True
#             for part in parts:
#                 if isinstance(ctx, dict) and part in ctx:
#                     ctx = ctx[part]
#                 else:
#                     valid_path = False
#                     break
#             if not valid_path:
#                 return False
#
#         return render_template_safe(jinja_env, template_str, context)
#     except Exception as e:
#         logger.debug(f"Error in render_template_bool: {str(e)}")
#         return False
#
#
# def safe_render_template_bool(jinja_env: Environment, template_str: str, context: Dict, default: bool = False) -> bool:
#     """
#     Safely render a template as boolean with a default fallback.
#
#     Args:
#         jinja_env: Jinja2 environment
#         template_str: Template string
#         context: Template context
#         default: Default value to return on error
#
#     Returns:
#         Boolean result or default on error
#     """
#     try:
#         result = render_template_safe(jinja_env, template_str, context)
#         if isinstance(result, bool):
#             return result
#         elif isinstance(result, str):
#             return result.lower() in ('true', '1', 'yes', 'on')
#         elif isinstance(result, (int, float)):
#             return bool(result)
#         else:
#             return default
#     except Exception as e:
#         logger.debug(f"Error in safe_render_template_bool: {str(e)}")
#         return default


def quote_jinja2_expressions(yaml_text: str) -> str:
    """
    Add quotes around Jinja2 expressions in YAML text.

    Args:
        yaml_text: YAML text content

    Returns:
        YAML text with quoted Jinja2 expressions
    """
    jinja_expr_pattern = re.compile(r'''
        ^(\s*[^:\n]+:\s*)         # YAML key and colon with optional indent
        (?!["'])                  # Not already quoted
        (.*{{.*}}.*?)             # Contains Jinja2 template
        (?<!["'])\s*$             # Not ending with a quote
    ''', re.VERBOSE)

    def replacer(match):
        key_part = match.group(1)
        value_part = match.group(2).strip()
        return f'{key_part}"{value_part}"'

    fixed_lines = []
    for line in yaml_text.splitlines():
        fixed_line = jinja_expr_pattern.sub(replacer, line)
        fixed_lines.append(fixed_line)
    return "\n".join(fixed_lines)



import os
import asyncio
import re
import sys
import json
import logging
import base64
import yaml
from collections import OrderedDict
from datetime import datetime, timedelta, timezone, date
import random
import string
from typing import Dict, Any, Optional, List, Union
import psycopg
from psycopg.rows import dict_row
from contextlib import contextmanager
try:
    from psycopg_pool import ConnectionPool
except ImportError:
    ConnectionPool = None

# logger = setup_logger(__name__, include_location=True)
#===================================
#  custom logger
#===================================

SUCCESS_LEVEL = 25
LOG_SEVERITY = {
    "DEBUG": "DEBUG",
    "INFO": "INFO",
    "SUCCESS": "SUCCESS",
    "WARNING": "WARNING",
    "ERROR": "ERROR",
    "CRITICAL": "CRITICAL"
}

LOG_COLORS = {
    "DEBUG": "\033[36m",      # Cyan
    "INFO": "\033[37m",       # White
    "SUCCESS": "\033[32m",    # Green
    "WARNING": "\033[33m",    # Yellow
    "ERROR": "\033[31m",      # Red
    "CRITICAL": "\033[35m"    # Magenta
}
RESET_COLOR = "\033[0m"

logging.addLevelName(SUCCESS_LEVEL, "SUCCESS")

class CustomLogger(logging.Logger):
    def success(self, message, *args, **kwargs):
        if self.isEnabledFor(SUCCESS_LEVEL):
            self.log(SUCCESS_LEVEL, message, *args, **kwargs, stacklevel=2)

def stringify_extra(value):
    if isinstance(value, (list, dict)):
        return str(value)
    else:
        return value


class CustomFormatter(logging.Formatter):

    def __init__(self, fmt="%(message)s", include_location=False, highlight_scope=True):
        super().__init__(fmt)
        self.include_location = include_location
        self.highlight_scope = highlight_scope

    def format(self, record):
        level_name = LOG_SEVERITY.get(record.levelname, record.levelname)
        level_color = LOG_COLORS.get(record.levelname, "")

        if hasattr(record, "scope"):
            scope_highlight = f"\033[32m{record.scope}\033[0m"
        else:
            scope_highlight = ""
        location = ""
        if self.include_location and hasattr(record, "module") and hasattr(record, "funcName") and hasattr(record,
                                                                                                           "lineno"):
            location = f"\033[1;33m({record.module}:{record.funcName}:{record.lineno})\033[0m"

        metadata_line = f"{level_color}[{level_name}]{RESET_COLOR} {scope_highlight} {location}".strip()

        if isinstance(record.msg, (dict, list)):
            message = str(record.msg)
        else:
            message = str(record.msg)

        message_split = message.splitlines()
        if len(message_split) > 1:
            message_line = f"     Message: {message_split[0]}"
            for line in message_split[1:]:
                message_line += f"\n             {line}"
        else:
            message_line = f"     Message: {message}" #.strip()

        extra_items = [
            f"{key}: {stringify_extra(value)}"
            for key, value in record.__dict__.items()
            if key not in [
                "msg", "args", "levelname", "levelno", "pathname",
                "filename", "module", "exc_info", "exc_text", "stack_info",
                "lineno", "funcName", "created", "msecs", "relativeCreated",
                "thread", "threadName", "process", "processName", "scope", "name", "taskName"
            ]
        ]
        extra_info = ""
        if extra_items:
            extra_info = f"\n     {' '.join(extra_items)}"
        formatted_log = f"{metadata_line}\n{message_line}{extra_info}"
        return formatted_log


class JSONFormatter(logging.Formatter):
    def format(self, record):
        log_dict = {
            "level": record.levelname,
            "message": record.msg,
            "time": self.formatTime(record, self.datefmt),
        }
        if hasattr(record, "scope"):
            log_dict["scope"] = record.scope
        if hasattr(record, "module") and hasattr(record, "funcName") and hasattr(record, "lineno"):
            log_dict["location"] = {
                "module": record.module,
                "function": record.funcName,
                "line": record.lineno,
            }
        return json.dumps(log_dict, ensure_ascii=False)

def setup_logger(name: str, include_location=False, use_json=False):
    logging.setLoggerClass(CustomLogger)
    logger = logging.getLogger(name)

    if not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setLevel(logging.DEBUG)
        if use_json:
            stream_handler.setFormatter(JSONFormatter())
        else:
            stream_handler.setFormatter(CustomFormatter(include_location=include_location))
        logger.addHandler(stream_handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False
    return logger



logger = setup_logger(__name__, include_location=True)

#===================================
#  time calendar
#===================================

def generate_id() -> str:
    start_date_time = datetime.now()
    timestamp = start_date_time.strftime("%Y%m%d_%H%M%S")
    microseconds = f"{start_date_time.microsecond:06d}"
    suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=4))
    return f"{timestamp}_{microseconds}_{suffix}"

def duration_seconds(start_date, end_date) -> int | None:
    return int((start_date - end_date).total_seconds())

def now_utc() -> datetime:
    return datetime.now(timezone.utc)

def today_utc() -> datetime:
    now = now_utc()
    return now.replace(hour=0, minute=0, second=0, microsecond=0)

def week_start_utc(reference: datetime = None) -> datetime:
    if reference is None:
        reference = now_utc()
    monday = reference - timedelta(days=reference.weekday())
    return monday.replace(hour=0, minute=0, second=0, microsecond=0)

def add_days(base: datetime, days: int) -> datetime:
    return base + timedelta(days=days)

def add_weeks(base: datetime, weeks: int) -> datetime:
    return base + timedelta(weeks=weeks)

def month_end(date: datetime) -> datetime:
    next_month = date.replace(day=28) + timedelta(days=4)
    last_day = next_month - timedelta(days=next_month.day)
    return last_day.replace(hour=23, minute=59, second=59, microsecond=999999)

def format_iso8601(dt: datetime) -> str:
    return dt.astimezone(timezone.utc).isoformat()

def format_utc(dt: datetime) -> str:
    return dt.strftime('%Y-%m-%d %H:%M:%S UTC')

def parse_iso8601(iso_str: str) -> datetime:
    return datetime.fromisoformat(iso_str)

def days_between(start: datetime, end: datetime) -> int:
    delta = end - start
    return delta.days

def is_weekend(date: datetime) -> bool:
    return date.weekday() >= 5

def next_weekday(date: datetime, weekday: int) -> datetime:
    days_ahead = weekday - date.weekday()
    if days_ahead <= 0:
        days_ahead += 7
    return date + timedelta(days=days_ahead)

#===================================
# environment
#===================================


def log_level() -> int:
    log_level: int = getattr(logging, os.environ.get("LOG_LEVEL", "INFO").upper(), logging.INFO)
    return log_level

def log_level_name() -> str:
    return logging.getLevelName(log_level())

def is_on(value: str) -> bool:
    return str(value).lower() in ("true", "1", "yes", "on", "y", "t")

def is_off(value: str) -> bool:
    return str(value).lower() in ("false", "0", "no", "off", "n", "f")

def get_bool(value) -> bool:
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        return value.strip().lower() in ("true", "1", "yes", "on", "pass", "ignore")
    if isinstance(value, (int, float)):
        return value != 0
    return False

async def mkdir(dir_path):
    try:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, lambda: os.makedirs(dir_path, exist_ok=True))
        return {f"Directory created": {dir_path}}
    except Exception as e:
        raise f"Failed to create directory: {dir_path}. Error: {e}"

#===================================
# serialization
#===================================


def make_serializable(value):
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, Exception):
        return str(value)
    if isinstance(value, dict):
        return {k: make_serializable(v) for k, v in value.items()}
    if isinstance(value, list):
        return [make_serializable(v) for v in value]
    if hasattr(value, "__dict__"):
        return {k: make_serializable(v) for k, v in value.__dict__.items()}
    return value


class SafeEncoder(json.JSONEncoder):
    def default(self, obj):
        return None

class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, date):
            return obj.isoformat()
        return super().default(obj)

def ordered_yaml_dump(data):
    def convert_ordered_dict(obj):
        if isinstance(obj, OrderedDict):
            return {k: convert_ordered_dict(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [convert_ordered_dict(i) for i in obj]
        return obj

    cleaned_data = convert_ordered_dict(data)
    return yaml.dump(cleaned_data, default_flow_style=False, sort_keys=False)

def ordered_yaml_load(stream):
    class OrderedLoader(yaml.SafeLoader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return OrderedDict(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping
    )
    return yaml.load(stream, OrderedLoader)

def encode_version(version: str) -> str:
    major, minor, patch = map(int, version.split("."))
    return f"{major:03d}.{minor:03d}.{patch:03d}"

def decode_version(encoded_version: str) -> str:
    major, minor, patch = encoded_version.split(".")
    return f"{int(major)}.{int(minor)}.{int(patch)}"

def increment_version(version: str) -> str:
    major, minor, patch = map(int, version.split("."))

    patch += 1
    if patch > 999:
        patch = 0
        minor += 1
        if minor > 999:
            minor = 0
            major += 1
            if major > 999:
                raise ValueError("Version overflow -> version limit is '999.999.999'")

    return f"{major}.{minor}.{patch}"

#===================================
# merging dictionaries and lists
#===================================

def deep_merge(dest: Union[Dict, List, Any], source: Union[Dict, List, Any]) -> Union[Dict, List, Any]:
    if isinstance(dest, dict) and isinstance(source, dict):
        result = dest.copy()
        for key, value in source.items():
            if key in result and isinstance(result[key], (dict, list)) and isinstance(value, (dict, list)):
                result[key] = deep_merge(result[key], value)
            else:
                result[key] = value
        return result
    elif isinstance(dest, list) and isinstance(source, list):
        if all(isinstance(item, dict) for item in dest + source):
            common_keys = set()
            for item in dest + source:
                common_keys.update(item.keys())
            if 'name' in common_keys:
                result = dest.copy()
                source_names = {item.get('name'): item for item in source if 'name' in item}
                for i, item in enumerate(result):
                    if 'name' in item and item['name'] in source_names:
                        result[i] = deep_merge(item, source_names[item['name']])
                        del source_names[item['name']]

                result.extend(source_names.values())
                return result

        result = dest.copy()
        result.extend(source)
        return result
    else:
        return source

#===================================
# connection strings
#===================================

def get_pgdb_connection(
    db_name: str = None,
    user: str = None,
    password: str = None,
    host: str = None,
    port: str = None,
    schema: str = None
) -> str:
    db_name = db_name or os.environ.get('POSTGRES_DB', 'noetl')
    user = user or os.environ.get('NOETL_USER', 'noetl')
    password = password or os.environ.get('NOETL_PASSWORD', 'noetl')
    host = host or os.environ.get('POSTGRES_HOST', 'localhost')
    port = port or os.environ.get('POSTGRES_PORT', '5432')
    schema = schema or os.environ.get('NOETL_SCHEMA', 'noetl')

    return f"dbname={db_name} user={user} password={password} host={host} port={port} hostaddr='' gssencmode=disable options='-c search_path={schema}'"

#===================================
# postgres pool
#===================================
db_pool = None

def initialize_db_pool():
    global db_pool
    if db_pool is None and ConnectionPool:
        try:
            db_pool = ConnectionPool(conninfo=get_pgdb_connection(), min_size=1, max_size=10)
            logger.info("Database connection pool initialized successfully")
        except Exception as e:
            logger.warning(f"Failed to initialize connection pool: {e}. Trying to use direct connections.")
            db_pool = None
    return db_pool

@contextmanager
def get_db_connection(optional=False):
    """
    Get a database connection from the pool or create a new one.
    
    Args:
        optional (bool): If True, return None instead of raising an exception when the connection fails.
                         This allows the application to continue without a database connection.
    """
    pool = initialize_db_pool()
    if pool:
        try:
            conn = pool.getconn()
            try:
                yield conn
                return
            finally:
                pool.putconn(conn)
        except Exception as pool_error:
            logger.warning(f"Connection pool error: {pool_error}. Falling back to direct connection.")

    conn = None
    try:
        conn = psycopg.connect(get_pgdb_connection())
        yield conn
    except Exception as e:
        logger.error(f"Connection failed: {e}")
        if optional:
            logger.warning("Database connection is optional, continuing without it.")
            yield None
        else:
            raise
    finally:
        if conn:
            conn.close()



import os
import json
import uuid
from typing import List, Optional
import psycopg
from noetl.common import setup_logger

logger = setup_logger(__name__, include_location=True)

class DatabaseSchema:

    def __init__(self, conn=None, pgdb: str = None, admin_conn: str = None, noetl_user: str = None, noetl_password: str = None, noetl_schema: str = None, auto_setup: bool = False):
        """
        Initialize the Database Schema.

        Args:
            conn: A database connection (will be created if None)
            pgdb: Postgres connection string for noetl user
            admin_conn: Admin connection string for creating user/schema
            noetl_user: Username for noetl operations (will be determined from env vars if None)
            noetl_password: Password for noetl user (will be determined from env vars if None)
            noetl_schema: Schema name for noetl tables (will be determined from env vars if None)
            auto_setup: If True, automatically create user/schema during initialization (default: False)
        """
        self.conn = conn
        self.pgdb = pgdb
        self.admin_conn = admin_conn
        self.auto_setup = auto_setup
        self.is_postgres = False
        self.admin_connection = None

        self.set_noetl_credentials(noetl_user, noetl_password, noetl_schema)

        if self.conn is None:
            self.initialize_connection()

    def set_noetl_credentials(self, noetl_user: str = None, noetl_password: str = None, noetl_schema: str = None):
        self.noetl_user = noetl_user or os.environ.get('NOETL_USER', 'noetl')
        self.noetl_password = noetl_password or os.environ.get('NOETL_PASSWORD', 'noetl')
        self.noetl_schema = noetl_schema or os.environ.get('NOETL_SCHEMA', 'noetl')
        logger.info(f"Using admin credentials for setup, NoETL user: {self.noetl_user}, schema: {self.noetl_schema}")


    def initialize_connection(self):
        try:
            if self.admin_conn is None:
                postgres_user = os.environ.get('POSTGRES_USER')
                postgres_password = os.environ.get('POSTGRES_PASSWORD')
                db_name = os.environ.get('POSTGRES_DB', 'postgres')
                host = os.environ.get('POSTGRES_HOST', 'localhost')
                port = os.environ.get('POSTGRES_PORT', '5432')
                self.admin_conn = f"dbname={db_name} user={postgres_user} password={postgres_password} host={host} port={port} hostaddr='' gssencmode=disable"
                logger.info(f"Using admin connection: dbname={db_name} user={postgres_user} host={host} port={port}")


            if self.pgdb is None:
                db_name = os.environ.get('POSTGRES_DB', 'postgres')
                host = os.environ.get('POSTGRES_HOST', 'localhost')
                port = os.environ.get('POSTGRES_PORT', '5432')

                self.pgdb = f"dbname={db_name} user={self.noetl_user} password={self.noetl_password} host={host} port={port} hostaddr='' gssencmode=disable"
                logger.info(f"NoETL connection: dbname={db_name} user={self.noetl_user} host={host} port={port}")

            if self.auto_setup:
                self.create_noetl_schema()

            try:
                self.conn = psycopg.connect(self.pgdb)
                self.is_postgres = True
                logger.info("Connected to Postgres database as noetl user.")
            except psycopg.OperationalError as e:
                postgres_user = os.environ.get('POSTGRES_USER')
                postgres_password = os.environ.get('POSTGRES_PASSWORD')

                if postgres_user and postgres_password:
                    logger.info(f"NoETL user connection failed, attempting to create user/schema: {e}")
                    self.create_noetl_schema()
                    self.conn = psycopg.connect(self.pgdb)
                    self.is_postgres = True
                    logger.info("Connected to Postgres database as noetl user after creating infrastructure.")
                else:
                    raise

        except Exception as e:
            logger.error(f"Failed to connect to Postgres: {e}.")
            raise

    def create_noetl_metadata(self):
        try:
            logger.info("Verifying NoETL.")
            try:
                test_conn = psycopg.connect(self.pgdb)
                test_conn.close()
                logger.info("NoETL user connected.")

                self.conn = psycopg.connect(self.pgdb)
                with self.conn.cursor() as cursor:
                    cursor.execute("""
                        SELECT 1 FROM information_schema.schemata WHERE schema_name = %s
                    """, (self.noetl_schema,))

                    if cursor.fetchone():
                        logger.info(f"NoETL schema '{self.noetl_schema}' exists.")
                        return True
                    else:
                        logger.info(f"NoETL schema '{self.noetl_schema}' does not exist, trying to create new one.")
                        self.conn.close()
                        self.create_noetl_schema()
                        return True

            except psycopg.OperationalError as e:
                logger.info(f"NoETL user does not exist or cannot connect: {e}")
                logger.info("Creating NoETL user and schema.")
                self.create_noetl_schema()
                return True

        except Exception as e:
            logger.error(f"Error verifying/setting up NoETL infrastructure: {e}.", exc_info=True)
            raise

    def create_noetl_schema(self):
        try:
            logger.info("Setting up noetl user and schema.")
            self.admin_connection = psycopg.connect(self.admin_conn)

            with self.admin_connection.cursor() as cursor:
                cursor.execute("""
                    SELECT 1 FROM pg_roles WHERE rolname = %s
                """, (self.noetl_user,))

                if not cursor.fetchone():
                    logger.info(f"Creating user '{self.noetl_user}'...")
                    create_user_sql = f"""
                        CREATE USER {self.noetl_user} WITH 
                        PASSWORD '{self.noetl_password}'
                        CREATEDB
                        LOGIN
                    """
                    cursor.execute(create_user_sql)
                    logger.info(f"User '{self.noetl_user}' created successfully.")
                else:
                    logger.info(f"User '{self.noetl_user}' already exists.")

                cursor.execute("""
                    SELECT 1 FROM information_schema.schemata WHERE schema_name = %s
                """, (self.noetl_schema,))

                if not cursor.fetchone():
                    logger.info(f"Creating schema '{self.noetl_schema}'.")
                    cursor.execute(f"CREATE SCHEMA {self.noetl_schema}")
                    logger.info(f"Schema '{self.noetl_schema}' created successfully.")
                else:
                    logger.info(f"Schema '{self.noetl_schema}' already exists.")

                logger.info(f"Granting permissions to user '{self.noetl_user}'.")
                cursor.execute(f"GRANT ALL PRIVILEGES ON SCHEMA {self.noetl_schema} TO {self.noetl_user}")
                cursor.execute(f"GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA {self.noetl_schema} TO {self.noetl_user}")
                cursor.execute(f"GRANT ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA {self.noetl_schema} TO {self.noetl_user}")
                cursor.execute(f"ALTER DEFAULT PRIVILEGES IN SCHEMA {self.noetl_schema} GRANT ALL ON TABLES TO {self.noetl_user}")
                cursor.execute(f"ALTER DEFAULT PRIVILEGES IN SCHEMA {self.noetl_schema} GRANT ALL ON SEQUENCES TO {self.noetl_user}")

                self.admin_connection.commit()
                logger.info("NoETL user and schema setup completed.")

        except Exception as e:
            logger.error(f"Error setting up noetl user and schema: {e}.", exc_info=True)
            if self.admin_connection:
                self.admin_connection.rollback()
            raise
        finally:
            if self.admin_connection:
                self.admin_connection.close()
                self.admin_connection = None

    def init_database(self):
        try:
            logger.info("Initializing database tables.")
            self.test_connection()
            self.set_search_path()
            self.create_postgres_tables()
            tables = self.list_tables()
            logger.info(f"Tables in noetl schema: {tables}.")

            return True
        except Exception as e:
            logger.error(f"Error initializing database: {e}.", exc_info=True)
            raise

    def set_search_path(self):
        """Set the search path to use the noetl schema"""
        try:
            with self.conn.cursor() as cursor:
                cursor.execute(f"SET search_path TO {self.noetl_schema}, public")
                self.conn.commit()
                logger.info(f"Search path set to '{self.noetl_schema}, public'")
        except Exception as e:
            logger.error(f"Error setting search path: {e}.", exc_info=True)
            raise

    def test_connection(self):
        try:
            with self.conn.cursor() as cursor:
                cursor.execute("SELECT 1")
                result = cursor.fetchone()
                if result and result[0] == 1:
                    logger.info("Database connected.")
                else:
                    logger.error("Database connection test failed.")
        except Exception as e:
            logger.error(f"Error testing database connection: {e}.", exc_info=True)
            raise

    def create_postgres_tables(self):
        with self.conn.cursor() as cursor:
            logger.info("Creating resource table.")
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.noetl_schema}.resource (
                    name TEXT PRIMARY KEY
                )
            """)

            logger.info("Creating catalog table.")
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.noetl_schema}.catalog (
                    resource_path     TEXT     NOT NULL,
                    resource_type     TEXT     NOT NULL REFERENCES {self.noetl_schema}.resource(name),
                    resource_version  TEXT     NOT NULL,
                    source            TEXT     NOT NULL DEFAULT 'inline',
                    resource_location TEXT,
                    content           TEXT,
                    payload           JSONB    NOT NULL,
                    meta              JSONB,
                    template          TEXT,
                    timestamp         TIMESTAMPTZ NOT NULL DEFAULT now(),
                    PRIMARY KEY (resource_path, resource_version)
                )
            """)

            logger.info("Creating workload table.")
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.noetl_schema}.workload (
                    execution_id VARCHAR,
                    timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    data TEXT,
                    PRIMARY KEY (execution_id)
                )
            """)
            self.test_workload_table()

            logger.info("Creating event_log table.")
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.noetl_schema}.event_log (
                    execution_id VARCHAR,
                    event_id VARCHAR,
                    parent_event_id VARCHAR,
                    timestamp TIMESTAMP,
                    event_type VARCHAR,
                    node_id VARCHAR,
                    node_name VARCHAR,
                    node_type VARCHAR,
                    status VARCHAR,
                    duration DOUBLE PRECISION,
                    input_context TEXT,
                    output_result TEXT,
                    metadata TEXT,
                    error TEXT,
                    loop_id VARCHAR,
                    loop_name VARCHAR,
                    iterator VARCHAR,
                    items TEXT,
                    current_index INTEGER,
                    current_item TEXT,
                    results TEXT,
                    worker_id VARCHAR,
                    distributed_state VARCHAR,
                    context_key VARCHAR,
                    context_value TEXT,
                    PRIMARY KEY (execution_id, event_id)
                )
            """)

            logger.info("Creating workflow table.")
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.noetl_schema}.workflow (
                    execution_id VARCHAR,
                    step_id VARCHAR,
                    step_name VARCHAR,
                    step_type VARCHAR,
                    description TEXT,
                    raw_config TEXT,
                    PRIMARY KEY (execution_id, step_id)
                )
            """)

            logger.info("Creating workbook table.")
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.noetl_schema}.workbook (
                    execution_id VARCHAR,
                    task_id VARCHAR,
                    task_name VARCHAR,
                    task_type VARCHAR,
                    raw_config TEXT,
                    PRIMARY KEY (execution_id, task_id)
                )
            """)

            logger.info("Creating transition table.")
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.noetl_schema}.transition (
                    execution_id VARCHAR,
                    from_step VARCHAR,
                    to_step VARCHAR,
                    condition TEXT,
                    with_params TEXT,
                    PRIMARY KEY (execution_id, from_step, to_step, condition)
                )
            """)

            self.conn.commit()
            logger.info("Postgres database tables initialized in noetl schema.")

    def test_workload_table(self):
        try:
            with self.conn.cursor() as cursor:
                cursor.execute(f"""
                    SELECT EXISTS (
                        SELECT FROM information_schema.tables 
                        WHERE table_schema = %s AND table_name = 'workload'
                    )
                """, (self.noetl_schema,))
                table_exists = cursor.fetchone()[0]
                if not table_exists:
                    logger.error("Failed to create workload table.")
                else:
                    logger.info("Workload table created.")

                    try:
                        test_id = f"test_{uuid.uuid4()}"
                        test_data = json.dumps({"test": "data"})
                        logger.info(f"Testing workload table with test_id: {test_id}.")

                        cursor.execute(f"""
                            INSERT INTO {self.noetl_schema}.workload (execution_id, data)
                            VALUES (%s, %s)
                            ON CONFLICT (execution_id) DO UPDATE
                            SET data = EXCLUDED.data
                        """, (test_id, test_data))
                        self.conn.commit()

                        cursor.execute(f"""
                            SELECT data FROM {self.noetl_schema}.workload WHERE execution_id = %s
                        """, (test_id,))
                        row = cursor.fetchone()
                        if row and row[0] == test_data:
                            logger.info("Tested workload table insert and select.")
                        else:
                            logger.error(f"Failed to verify test data in workload table. Expected: {test_data}, Got: {row[0] if row else None}")

                        cursor.execute(f"""
                            DELETE FROM {self.noetl_schema}.workload WHERE execution_id = %s
                        """, (test_id,))
                        self.conn.commit()
                        logger.info("Cleaned up test data from workload table.")
                    except Exception as e:
                        logger.error(f"Error testing workload table: {e}.", exc_info=True)
        except Exception as e:
            logger.error(f"Error testing workload table: {e}.", exc_info=True)

    def list_tables(self) -> List[str]:
        try:
            with self.conn.cursor() as cursor:
                cursor.execute("""
                    SELECT table_name 
                    FROM information_schema.tables 
                    WHERE table_schema = %s
                """, (self.noetl_schema,))
                return [row[0] for row in cursor.fetchall()]
        except Exception as e:
            logger.error(f"Error listing tables: {e}.", exc_info=True)
            return []

    def table_exists(self, table_name: str) -> bool:
        try:
            with self.conn.cursor() as cursor:
                cursor.execute("""
                    SELECT EXISTS (
                        SELECT FROM information_schema.tables 
                        WHERE table_schema = %s AND table_name = %s
                    )
                """, (self.noetl_schema, table_name))
                return cursor.fetchone()[0]
        except Exception as e:
            logger.error(f"Error checking if table exists: {e}.", exc_info=True)
            return False

    def drop_noetl_schema(self):
        """Drop the entire noetl schema and all its objects"""
        try:
            logger.warning(f"Dropping schema '{self.noetl_schema}' and all its objects.")
            self.admin_connection = psycopg.connect(self.admin_conn)

            with self.admin_connection.cursor() as cursor:
                cursor.execute(f"DROP SCHEMA IF EXISTS {self.noetl_schema} CASCADE")
                self.admin_connection.commit()
                logger.info(f"Schema '{self.noetl_schema}' dropped.")

        except Exception as e:
            logger.error(f"Error dropping schema: {e}.", exc_info=True)
            if self.admin_connection:
                self.admin_connection.rollback()
            raise
        finally:
            if self.admin_connection:
                self.admin_connection.close()
                self.admin_connection = None

    def close(self):
        if self.conn:
            try:
                self.conn.close()
                logger.info("Database connection closed.")
            except Exception as e:
                logger.error(f"Error closing database connection: {e}.", exc_info=True)



import os
import sys
import unittest
from unittest.mock import Mock, patch
import json
import yaml
import base64
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from noetl.keyval import KeyVal, KeyValBuilder

class TestKeyVal(unittest.TestCase):
    def setUp(self):
        self.test_data = {
            'name': 'test',
            'version': '1.0.0',
            'config': {
                'database': {
                    'host': 'localhost',
                    'port': 5432
                },
                'api': {
                    'timeout': 30,
                    'retries': 3
                }
            },
            'items': ['item1', 'item2', 'item3']
        }
        self.keyval = KeyVal(self.test_data)

    def test_keyval_initialization(self):
        kv = KeyVal()
        self.assertIsInstance(kv, dict)
        self.assertIsInstance(kv, KeyVal)

        kv_with_data = KeyVal(self.test_data)
        self.assertEqual(kv_with_data['name'], 'test')

    def test_builder_pattern(self):
        builder = KeyVal.builder()
        self.assertIsInstance(builder, KeyValBuilder)

    def test_info_property(self):
        kv = KeyVal()
        self.assertIsNone(kv.info)

        test_info = {'type': 'test', 'description': 'test keyval'}
        kv.info = test_info
        self.assertEqual(kv.info, test_info)

    def test_get_keys_root_level(self):
        keys = self.keyval.get_keys()
        expected_keys = ['name', 'version', 'config', 'items']
        self.assertEqual(sorted(keys), sorted(expected_keys))

    def test_get_keys_nested_path(self):
        keys = self.keyval.get_keys('config')
        expected_keys = ['config.database', 'config.api']
        self.assertEqual(sorted(keys), sorted(expected_keys))

    def test_get_keys_deeper_nested_path(self):
        keys = self.keyval.get_keys('config.database')
        expected_keys = ['config.database.host', 'config.database.port']
        self.assertEqual(sorted(keys), sorted(expected_keys))

    def test_get_keys_invalid_path(self):
        keys = self.keyval.get_keys('nonexistent')
        self.assertEqual(keys, [])

    def test_get_value_simple_key(self):
        value = self.keyval.get_value('name')
        self.assertEqual(value, 'test')

    def test_get_value_nested_key(self):
        value = self.keyval.get_value('config.database.host')
        self.assertEqual(value, 'localhost')

    def test_get_value_with_default(self):
        value = self.keyval.get_value('nonexistent', default='default_value')
        self.assertEqual(value, 'default_value')

    def test_get_value_no_path(self):
        value = self.keyval.get_value()
        self.assertEqual(value, self.test_data)

    def test_get_value_with_exclude(self):
        value = self.keyval.get_value(exclude=['version'])
        self.assertNotIn('version', value)
        self.assertIn('name', value)

    def test_set_value(self):
        kv = KeyVal()
        kv.set_value("a.b.c", 10)
        self.assertEqual(kv.get_value("a.b.c"), 10)

        with self.assertRaises(ValueError):
            kv.set_value("", 10)

        with self.assertRaises(ValueError):
            kv.set_value(None, 10)

    def test_delete_value(self):
        kv = KeyVal({"a": {"b": {"c": 10}}})
        kv.delete_value("a.b.c")
        self.assertIsNone(kv.get_value("a.b.c"))

    def test_delete_keys(self):
        kv = KeyVal({"a": 1, "b": 2, "c": 3})
        kv.delete_keys(["a", "c"])
        self.assertEqual(kv, {"b": 2})

    def test_retain_keys(self):
        kv = KeyVal({"a": 1, "b": 2, "c": 3})
        kv.retain_keys(["a", "c"])
        self.assertEqual(sorted(kv.keys()), ["a", "c"])

    def test_add_method(self):
        kv = KeyVal()
        result = kv.add("test.key", "value")
        self.assertIs(result, kv)
        self.assertEqual(kv.get_value("test.key"), "value")

    def test_to_json(self):
        kv = KeyVal({"a": 1, "b": 2})
        json_repr = kv.to_json()
        self.assertIsInstance(json_repr, bytes)
        self.assertEqual(json.loads(json_repr.decode('utf-8')), {"a": 1, "b": 2})

    def test_as_json(self):
        kv = KeyVal({'a': 1, 'b': 2})
        self.assertEqual(kv.as_json(path='a'), '1')
        self.assertEqual(kv.as_json(path='b'), '2')
        self.assertEqual(kv.as_json(indent=2), '{\n  "a": 1,\n  "b": 2\n}')

        with self.assertRaises(ValueError):
            kv.as_json(path='nonexistent')

    def test_get_keyval(self):
        kv = KeyVal({"nested": {"key": "value"}})
        nested_kv = kv.get_keyval("nested")
        self.assertIsInstance(nested_kv, KeyVal)
        self.assertEqual(nested_kv["key"], "value")

    def test_encode_decode(self):
        kv = KeyVal({"a": 1, "b": 2})
        encoded = kv.encode()
        self.assertIsInstance(encoded, bytes)

        decoded = KeyVal.decode(encoded)
        self.assertEqual(decoded, {"a": 1, "b": 2})

    def test_encode_with_keys(self):
        kv = KeyVal({"a": 1, "b": 2, "c": 3})
        encoded = kv.encode(keys=["a", "c"])
        decoded = KeyVal.decode(encoded)
        self.assertEqual(decoded, {"a": 1, "c": 3})

    def test_str_to_base64(self):
        result = KeyVal.str_to_base64("test")
        expected = base64.b64encode("test".encode()).decode()
        self.assertEqual(result, expected)

    def test_base64_to_str(self):
        encoded = base64.b64encode("test".encode()).decode()
        result = KeyVal.base64_to_str(encoded)
        self.assertEqual(result, "test")

    def test_base64_to_yaml(self):
        data = {"key": "value"}
        yaml_str = yaml.safe_dump(data)
        encoded = base64.b64encode(yaml_str.encode()).decode()
        result = KeyVal.base64_to_yaml(encoded)
        self.assertEqual(result, data)

    def test_yaml_dump(self):
        data = {"key": "value"}
        result = KeyVal.yaml_dump(data)
        self.assertIsInstance(result, str)
        self.assertEqual(yaml.safe_load(result), data)

    def test_base64_value(self):
        kv = KeyVal({"workflow_base64": "test"})
        result = kv.base64_value("workflow_base64")
        expected = base64.b64encode("test".encode()).decode()
        self.assertEqual(result, expected)

        with self.assertRaises(ValueError):
            kv.base64_value("nonexistent")

    def test_yaml_value(self):
        data = {"key": "value"}
        yaml_str = yaml.safe_dump(data)
        encoded = base64.b64encode(yaml_str.encode()).decode()
        kv = KeyVal({"value": encoded})
        result = kv.yaml_value()
        self.assertEqual(result, data)

    def test_yaml_value_dump(self):
        data = {"key": "value"}
        yaml_str = yaml.safe_dump(data)
        encoded = base64.b64encode(yaml_str.encode()).decode()
        kv = KeyVal({"value": encoded})
        result = kv.yaml_value_dump()
        self.assertIsInstance(result, str)


class TestKeyValBuilder(unittest.TestCase):
    def test_builder_add(self):
        builder = KeyValBuilder()
        result = builder.add("test.key", "value")
        self.assertIs(result, builder)

    def test_builder_remove(self):
        builder = KeyValBuilder()
        builder.add("test.key", "value")
        result = builder.remove("test.key")
        self.assertIs(result, builder)

    def test_builder_info(self):
        builder = KeyValBuilder()
        metadata = {"created_by": "test"}
        result = builder.info(metadata)
        self.assertIs(result, builder)

    def test_builder_build(self):
        kv = (KeyValBuilder()
              .add("user.name", "test")
              .add("user.age", 25)
              .info({"created_by": "test"})
              .build())

        self.assertIsInstance(kv, KeyVal)
        self.assertEqual(kv.get_value("user.name"), "test")
        self.assertEqual(kv.get_value("user.age"), 25)
        self.assertEqual(kv.info, {"created_by": "test"})


if __name__ == '__main__':
    unittest.main()



import os
import re
import uuid
import datetime
import httpx
import psycopg
import json
from typing import Dict, Any
from jinja2 import Environment
from noetl.common import DateTimeEncoder, setup_logger
from noetl.render import render_template

try:
    import duckdb
    DUCKDB_AVAILABLE = True
except ImportError:
    DUCKDB_AVAILABLE = False

logger = setup_logger(__name__, include_location=True)

import threading
from contextlib import contextmanager

_duckdb_connections = {}
_connection_lock = threading.Lock()

@contextmanager
def get_duckdb_connection(duckdb_file_path):
    """Context manager for shared DuckDB connections to maintain attachments"""
    logger.debug("=== ACTION.GET_DUCKDB_CONNECTION: Function entry ===")
    logger.debug(f"ACTION.GET_DUCKDB_CONNECTION: duckdb_file_path={duckdb_file_path}")

    with _connection_lock:
        if duckdb_file_path not in _duckdb_connections:
            logger.debug(f"ACTION.GET_DUCKDB_CONNECTION: Creating new DuckDB connection for {duckdb_file_path}")
            _duckdb_connections[duckdb_file_path] = duckdb.connect(duckdb_file_path)
        else:
            logger.debug(f"ACTION.GET_DUCKDB_CONNECTION: Reusing existing DuckDB connection for {duckdb_file_path}")
        conn = _duckdb_connections[duckdb_file_path]
    
    try:
        logger.debug("ACTION.GET_DUCKDB_CONNECTION: Yielding connection")
        yield conn
    finally:
        pass

def execute_http_task(task_config: Dict, context: Dict, jinja_env: Environment, task_with: Dict,
                      log_event_callback=None) -> Dict:
    """
    Execute an HTTP task.

    Args:
        task_config: The task configuration
        context: The context to use for rendering templates
        jinja_env: The Jinja2 environment for template rendering
        task_with: The rendered 'with' parameters dictionary
        log_event_callback: A callback function to log events

    Returns:
        A dictionary of the task result
    """
    logger.debug("=== ACTION.EXECUTE_HTTP_TASK: Function entry ===")
    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Parameters - task_config={task_config}, task_with={task_with}")

    task_id = str(uuid.uuid4())
    task_name = task_config.get('task', 'http_task')
    start_time = datetime.datetime.now()

    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Generated task_id={task_id}")
    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Task name={task_name}")
    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Start time={start_time.isoformat()}")

    try:
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Rendering HTTP task configuration")
        method = task_config.get('method', 'GET').upper()
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: HTTP method={method}")

        endpoint = render_template(jinja_env, task_config.get('endpoint', ''), context)
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Rendered endpoint={endpoint}")

        params = render_template(jinja_env, task_config.get('params', {}), context)
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Rendered params={params}")

        payload = render_template(jinja_env, task_config.get('payload', {}), context)
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Rendered payload={payload}")

        headers = render_template(jinja_env, task_config.get('headers', {}), context)
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Rendered headers={headers}")

        timeout = task_config.get('timeout', 30)
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Timeout={timeout}")

        logger.info(f"ACTION.EXECUTE_HTTP_TASK: Executing HTTP {method} request to {endpoint}")

        event_id = None
        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Writing task_start event log")
            event_id = log_event_callback(
                'task_start', task_id, task_name, 'http',
                'in_progress', 0, context, None,
                {'method': method, 'endpoint': endpoint, 'with_params': task_with}, None
            )

        headers = render_template(jinja_env, task_config.get('headers', {}), context)
        timeout = task_config.get('timeout', 30)

        try:
            logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Creating HTTP client with timeout={timeout}")
            with httpx.Client(timeout=timeout) as client:
                request_args = {
                    'url': endpoint,
                    'headers': headers,
                    'params': params
                }
                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Initial request_args={request_args}")

                if method in ['POST', 'PUT', 'PATCH'] and payload:
                    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Processing payload for {method} request")
                    content_type = headers.get('Content-Type', '').lower()
                    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Content-Type={content_type}")

                    if 'application/json' in content_type:
                        request_args['json'] = payload
                        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Using JSON payload")
                    elif 'application/x-www-form-urlencoded' in content_type:
                        request_args['data'] = payload
                        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Using form data payload")
                    elif 'multipart/form-data' in content_type:
                        request_args['files'] = payload
                        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Using multipart form data payload")
                    else:
                        if isinstance(payload, (dict, list)):
                            request_args['json'] = payload
                            logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Using default JSON payload for dict/list")
                        else:
                            request_args['data'] = payload
                            logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Using data payload for non-dict/list")

                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Final request_args={request_args}")
                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Making HTTP request")
                response = client.request(method, **request_args)
                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: HTTP response received - status_code={response.status_code}")

                response_data = {
                    'status_code': response.status_code,
                    'headers': dict(response.headers),
                    'url': str(response.url),
                    'elapsed': response.elapsed.total_seconds() if hasattr(response, 'elapsed') else None
                }
                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Response metadata={response_data}")

                try:
                    response_content_type = response.headers.get('Content-Type', '').lower()
                    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Response Content-Type={response_content_type}")

                    if 'application/json' in response_content_type:
                        response_data['data'] = response.json()
                        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Parsed JSON response data")
                    else:
                        response_data['data'] = response.text
                        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Using text response data")
                except Exception as e:
                    logger.warning(f"ACTION.EXECUTE_HTTP_TASK: Failed to parse response content: {str(e)}")
                    response_data['data'] = response.text

                is_success = response.is_success
                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Request success status={is_success}")

                result = {
                    'id': task_id,
                    'status': 'success' if is_success else 'error',
                    'data': response_data
                }

                if not is_success:
                    result['error'] = f"HTTP {response.status_code}: {response.reason_phrase}"
                    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Request failed with error={result['error']}")

                end_time = datetime.datetime.now()
                duration = (end_time - start_time).total_seconds()
                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Task duration={duration} seconds")

                if log_event_callback:
                    logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Writing task_complete event log")
                    log_event_callback(
                        'task_complete', task_id, task_name, 'http',
                        result['status'], duration, context, result.get('data'),
                        {'method': method, 'endpoint': endpoint, 'with_params': task_with}, event_id
                    )

                logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Returning result={result}")
                logger.debug("=== ACTION.EXECUTE_HTTP_TASK: Function exit (success) ===")
                return result

        except httpx.HTTPStatusError as e:
            error_msg = f"HTTP error: {e.response.status_code} - {e.response.text}"
            logger.error(f"ACTION.EXECUTE_HTTP_TASK: HTTPStatusError - {error_msg}")
            raise Exception(error_msg)
        except httpx.RequestError as e:
            error_msg = f"Request error: {str(e)}"
            logger.error(f"ACTION.EXECUTE_HTTP_TASK: RequestError - {error_msg}")
            raise Exception(error_msg)
        except httpx.TimeoutException as e:
            error_msg = f"Request timeout: {str(e)}"
            logger.error(f"ACTION.EXECUTE_HTTP_TASK: TimeoutException - {error_msg}")
            raise Exception(error_msg)

    except Exception as e:
        error_msg = str(e)
        logger.error(f"ACTION.EXECUTE_HTTP_TASK: Exception - {error_msg}", exc_info=True)
        end_time = datetime.datetime.now()
        duration = (end_time - start_time).total_seconds()
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Task duration={duration} seconds (error path)")

        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Writing task_error event log")
            log_event_callback(
                'task_error', task_id, task_name, 'http',
                'error', duration, context, None,
                {'error': error_msg, 'with_params': task_with}, event_id
            )

        result = {
            'id': task_id,
            'status': 'error',
            'error': error_msg
        }
        logger.debug(f"ACTION.EXECUTE_HTTP_TASK: Returning error result={result}")
        logger.debug("=== ACTION.EXECUTE_HTTP_TASK: Function exit (error) ===")
        return result

def execute_python_task(task_config: Dict, context: Dict, jinja_env: Environment, task_with: Dict, log_event_callback=None) -> Dict:
    """
    Execute a Python task.

    Args:
        task_config: The task configuration
        context: The context for rendering templates
        jinja_env: The Jinja2 environment for template rendering
        task_with: The rendered 'with' parameters dictionary
        log_event_callback: A callback function to log events

    Returns:
        A dictionary of the task result
    """
    logger.debug("=== ACTION.EXECUTE_PYTHON_TASK: Function entry ===")
    logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Parameters - task_config={task_config}, task_with={task_with}")

    task_id = str(uuid.uuid4())
    task_name = task_config.get('task', 'python_task')
    start_time = datetime.datetime.now()

    logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Generated task_id={task_id}")
    logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Task name={task_name}")
    logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Start time={start_time.isoformat()}")

    try:
        code = task_config.get('code', '')
        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Python code length={len(code)} chars")

        event_id = None
        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Writing task_start event log")
            event_id = log_event_callback(
                'task_start', task_id, task_name, 'python',
                'in_progress', 0, context, None,
                {'with_params': task_with}, None
            )
            logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Task start event_id={event_id}")

        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Setting up execution globals")
        exec_globals = {
            '__builtins__': __builtins__,
            'context': context,
            'os': os,
            'json': json,
            'datetime': datetime,
            'uuid': uuid
        }
        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Execution globals keys: {list(exec_globals.keys())}")

        exec_locals = {}
        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Executing Python code")
        exec(code, exec_globals, exec_locals)
        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Python execution completed")
        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Execution locals keys: {list(exec_locals.keys())}")

        if 'main' in exec_locals:
            result_data = exec_locals['main'](**task_with)
            end_time = datetime.datetime.now()
            duration = (end_time - start_time).total_seconds()

            if log_event_callback:
                log_event_callback(
                    'task_complete', task_id, task_name, 'python',
                    'success', duration, context, result_data,
                    {'with_params': task_with}, event_id
                )

            return {
                'id': task_id,
                'status': 'success',
                'data': result_data
            }
        else:
            error_msg = "Main function must be defined in Python task."
            end_time = datetime.datetime.now()
            duration = (end_time - start_time).total_seconds()

            if log_event_callback:
                log_event_callback(
                    'task_error', task_id, task_name, 'python',
                    'error', duration, context, None,
                    {'error': error_msg, 'with_params': task_with}, event_id
                )

            return {
                'id': task_id,
                'status': 'error',
                'error': error_msg
            }

    except Exception as e:
        error_msg = str(e)
        logger.error(f"ACTION.EXECUTE_PYTHON_TASK: Exception - {error_msg}", exc_info=True)

        end_time = datetime.datetime.now()
        duration = (end_time - start_time).total_seconds()
        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Task duration={duration} seconds (error path)")

        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Writing task_error event log")
            log_event_callback(
                'task_error', task_id, task_name, 'python',
                'error', duration, context, None,
                {'error': error_msg, 'with_params': task_with}, event_id
            )

        result = {
            'id': task_id,
            'status': 'error',
            'error': error_msg
        }
        logger.debug(f"ACTION.EXECUTE_PYTHON_TASK: Returning error result={result}")
        logger.debug("=== ACTION.EXECUTE_PYTHON_TASK: Function exit (error) ===")
        return result


def sql_split(sql_text):
    commands = []
    current_command = []
    in_single_quote = False
    in_double_quote = False
    i = 0

    while i < len(sql_text):
        char = sql_text[i]

        if char == "'" and not in_double_quote:
            if i > 0 and sql_text[i - 1] == '\\':
                pass
            else:
                in_single_quote = not in_single_quote
        elif char == '"' and not in_single_quote:
            if i > 0 and sql_text[i - 1] == '\\':
                pass
            else:
                in_double_quote = not in_double_quote
        elif char == ';' and not in_single_quote and not in_double_quote:
            cmd = ''.join(current_command).strip()
            if cmd:
                commands.append(cmd)
            current_command = []
            i += 1
            continue

        current_command.append(char)
        i += 1

    cmd = ''.join(current_command).strip()
    if cmd:
        commands.append(cmd)

    return commands



def execute_duckdb_task(task_config: Dict, context: Dict, jinja_env: Environment, task_with: Dict, log_event_callback=None) -> Dict:
    """
    Execute a DuckDB task.

    Args:
        task_config: The task configuration
        context: The context for rendering templates
        jinja_env: The Jinja2 environment for template rendering
        task_with: The rendered 'with' parameters dictionary
        log_event_callback: A callback function to log events

    Returns:
        A dictionary of the task result
    """
    logger.debug("=== ACTION.EXECUTE_DUCKDB_TASK: Function entry ===")
    logger.debug(f"ACTION.EXECUTE_DUCKDB_TASK: Parameters - task_config={task_config}, task_with={task_with}")

    if not DUCKDB_AVAILABLE:
        task_id = str(uuid.uuid4())
        task_name = task_config.get('task', 'duckdb_task')
        error_msg = "DuckDB is not installed. Install it with 'pip install noetl[duckdb]'."
        logger.error(error_msg)

        if log_event_callback:
            log_event_callback(
                'task_error', task_id, task_name, 'duckdb',
                'error', 0, context, None,
                {'error': error_msg, 'with_params': task_with}, None
            )

        return {
            'id': task_id,
            'status': 'error',
            'error': error_msg
        }

    task_id = str(uuid.uuid4())
    task_name = task_config.get('task', 'duckdb_task')
    start_time = datetime.datetime.now()

    logger.debug(f"ACTION.EXECUTE_DUCKDB_TASK: Generated task_id={task_id}")
    logger.debug(f"ACTION.EXECUTE_DUCKDB_TASK: Task name={task_name}")
    logger.debug(f"ACTION.EXECUTE_DUCKDB_TASK: Start time={start_time.isoformat()}")

    try:
        commands = task_config.get('command', task_config.get('commands', []))

        if isinstance(commands, str):
            commands_rendered = render_template(jinja_env, commands, {**context, **task_with})
            cmd_lines = []
            for line in commands_rendered.split('\n'):
                line = line.strip()
                if line and not line.startswith('--'):
                    cmd_lines.append(line)
            commands_text = ' '.join(cmd_lines)
            commands = sql_split(commands_text)
            # commands = [cmd.strip() for cmd in commands_text.split(';') if cmd.strip()]

        bucket = task_with.get('bucket', context.get('bucket', ''))
        blob_path = task_with.get('blob', '')
        file_path = task_with.get('file', '')
        table = task_with.get('table', '')

        event_id = None
        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_DUCKDB_TASK: Writing task_start event log")
            event_id = log_event_callback(
                'task_start', task_id, task_name, 'duckdb',
                'in_progress', 0, context, None,
                {'with_params': task_with}, None
            )

        import time
        duckdb_data_dir = os.environ.get("NOETL_DATA_DIR", "./data")
        execution_id = context.get("execution_id") or context.get("jobId") or (context.get("job", {}).get("uuid") if isinstance(context.get("job"), dict) else None) or "default"
        if isinstance(execution_id, str) and ('{{' in execution_id or '}}' in execution_id):
            execution_id = render_template(jinja_env, execution_id, context)
        duckdb_file = os.path.join(duckdb_data_dir, "noetldb", f"duckdb_{execution_id}.duckdb")
        os.makedirs(os.path.dirname(duckdb_file), exist_ok=True)
        logger.info(f"Connecting to DuckDB at {duckdb_file} for execution {execution_id}")
        duckdb_con = duckdb.connect(duckdb_file)

        db_type = task_with.get('db_type', 'postgres')
        db_alias = task_with.get('db_alias', 'postgres_db')

        if db_type.lower() == 'postgres':
            logger.info("Installing and loading Postgres extension")
            duckdb_con.execute("INSTALL postgres;")
            duckdb_con.execute("LOAD postgres;")
        elif db_type.lower() == 'mysql':
            logger.info("Installing and loading MySQL extension")
            duckdb_con.execute("INSTALL mysql;")
            duckdb_con.execute("LOAD mysql;")
        elif db_type.lower() == 'sqlite':
            logger.info("SQLite support is built-in to DuckDB, no extension needed")
        else:
            logger.info(f"Using custom database type: {db_type}, no specific extension loaded")

        duckdb_con.execute("INSTALL httpfs;")
        duckdb_con.execute("LOAD httpfs;")

        key_id = task_with.get('key_id')
        secret_key = task_with.get('secret_key')
        if key_id and secret_key:
            logger.info("Setting up S3 credentials for GCS operations")
            try:
                duckdb_con.execute(f"""
                    CREATE OR REPLACE SECRET gcs_secret (
                        TYPE S3,
                        PROVIDER GCS,
                        KEY_ID '{key_id}',
                        SECRET '{secret_key}',
                        REGION 'auto',
                        ENDPOINT 'storage.googleapis.com',
                        URL_STYLE 'path',
                        USE_SSL true
                    );
                """)
                logger.info("Successfully created GCS configuration chain")
                try:
                    secrets_list = duckdb_con.execute("SELECT * FROM duckdb_secrets();").fetchall()
                    gcs_secret_exists = any(secret[0] == 'gcs_secret' for secret in secrets_list)

                    if not gcs_secret_exists:
                        logger.info("Creating persistent GCS secret")
                        duckdb_con.execute(f"CREATE PERSISTENT SECRET gcs_secret (TYPE S3, KEY_ID '{key_id}', SECRET '{secret_key}');")
                except Exception as e:
                    logger.warning(f"Error checking or creating GCS secret: {e}. Will continue with chain configuration.")
            except Exception as e:
                logger.warning(f"Error creating GCS chain: {e}. Falling back to individual parameter configuration.")
                duckdb_con.execute("set s3_endpoint='storage.googleapis.com';")
                duckdb_con.execute("set s3_region='auto';")
                duckdb_con.execute("set s3_url_style='path';")
                duckdb_con.execute("set s3_use_ssl=true;")
                duckdb_con.execute(f"set s3_access_key_id='{key_id}';")
                duckdb_con.execute(f"set s3_secret_access_key='{secret_key}';")
                try:
                    secrets_list = duckdb_con.execute("SELECT * FROM duckdb_secrets();").fetchall()
                    gcs_secret_exists = any(secret[0] == 'gcs_secret' for secret in secrets_list)

                    if not gcs_secret_exists:
                        logger.info("Creating persistent GCS secret")
                        duckdb_con.execute(f"CREATE PERSISTENT SECRET gcs_secret (TYPE S3, KEY_ID '{key_id}', SECRET '{secret_key}');")
                except Exception as secret_e:
                    logger.warning(f"Error checking or creating GCS secret: {secret_e}. Will continue with session credentials.")

        if db_type.lower() == 'postgres':
            pg_host = task_with.get('db_host', os.environ.get('POSTGRES_HOST', 'localhost'))
            pg_port = task_with.get('db_port', os.environ.get('POSTGRES_PORT', '5434'))
            pg_user = task_with.get('db_user', os.environ.get('POSTGRES_USER', 'noetl'))
            pg_password = task_with.get('db_password', os.environ.get('POSTGRES_PASSWORD', 'noetl'))
            pg_db = task_with.get('db_name', os.environ.get('POSTGRES_DB', 'noetl'))
            pg_conn_string = f"dbname={pg_db} user={pg_user} password={pg_password} host={pg_host} port={pg_port}"
            conn_string = pg_conn_string
        elif db_type.lower() == 'sqlite':
            sqlite_path = task_with.get('db_path', os.path.join(duckdb_data_dir, 'sqlite', 'noetl.db'))
            conn_string = sqlite_path
        elif db_type.lower() == 'mysql':
            mysql_host = task_with.get('db_host', os.environ.get('MYSQL_HOST', 'localhost'))
            mysql_port = task_with.get('db_port', os.environ.get('MYSQL_PORT', '3306'))
            mysql_user = task_with.get('db_user', os.environ.get('MYSQL_USER', 'noetl'))
            mysql_password = task_with.get('db_password', os.environ.get('MYSQL_PASSWORD', 'noetl'))
            mysql_db = task_with.get('db_name', os.environ.get('MYSQL_DB', 'noetl'))
            mysql_conn_string = f"host={mysql_host} port={mysql_port} user={mysql_user} password={mysql_password} dbname={mysql_db}"
            conn_string = mysql_conn_string
        else:
            conn_string = task_with.get('db_conn_string', '')
            if not conn_string:
                logger.warning(f"No connection string provided for database type: {db_type}. Using in-memory DuckDB.")
                conn_string = "memory"

        read_only = task_with.get('db_read_only', False)
        attach_options = ""
        if read_only:
            attach_options += " (READ_ONLY)"
        elif db_type.lower() in ['postgres', 'mysql']:
            attach_options += f" (TYPE {db_type.lower()})"

        try:
            test_query = f"SELECT 1 FROM {db_alias}.sqlite_master LIMIT 1" if db_type.lower() == 'sqlite' else f"SELECT 1 FROM {db_alias}.information_schema.tables LIMIT 1"
            duckdb_con.execute(test_query)
            logger.info(f"Database '{db_alias}' is already attached.")
        except Exception as e:
            try:
                logger.info(f"Attaching {db_type} database as '{db_alias}'.")
                attach_sql = f"ATTACH '{conn_string}' AS {db_alias}{attach_options};"
                logger.debug(f"ATTACH SQL: {attach_sql}")
                duckdb_con.execute(attach_sql)
            except Exception as attach_error:
                logger.error(f"Error attaching database: {attach_error}.")
                raise

        results = {}
        if commands:
            for i, cmd in enumerate(commands):
                if isinstance(cmd, str) and ('{{' in cmd or '}}' in cmd):
                    cmd = render_template(jinja_env, cmd, context)
                    if "CREATE SECRET" in cmd or "CREATE OR REPLACE CHAIN" in cmd:
                        cmd = re.sub(r"{{[^}]*\|\s*default\(['\"]([^'\"]*)['\"].*?}}", r"\1", cmd)
                        cmd = re.sub(r"default\('([^']*)'\)", r'default("\1")', cmd)
                        cmd = re.sub(r"(HOST|DATABASE|USER|PASSWORD|ENDPOINT|REGION|URL_STYLE|KEY_ID|SECRET_KEY) '([^']*)'", r'\1 "\2"', cmd)

                logger.info(f"Executing DuckDB command: {cmd}")

                # very special case fordecimal separator in CAST operations
                # users need to avoid use that and istead duckdb support, and we support rendering like
                # , 'max_delta_t_d': 'NUMERIC'
                # , 'max_delta_aggressor': 'NUMERIC'
                # }
                # , header = { % if header %}true
                # { % else %}false
                # { % endif %}
                # , delim = '{{ delim }}'
                # , decimal_separator = '{{ decimal_separator }}'
                decimal_separator = task_with.get('decimal_separator')
                if decimal_separator and decimal_separator != '.':
                    cast_pattern = r'CAST\s*\(\s*([^\s]+)\s+AS\s+NUMERIC[^)]*\)'
                    def replace_cast(match):
                        column = match.group(1)
                        return f"CAST(REPLACE({column}, '{decimal_separator}', '.') AS NUMERIC)"
                    cmd = re.sub(cast_pattern, replace_cast, cmd, flags=re.IGNORECASE)
                
                if cmd.strip().upper().startswith("ATTACH"):
                    try:
                        attach_parts = cmd.strip().split(" AS ")
                        if len(attach_parts) >= 2:
                            db_alias_with_options = attach_parts[1].strip()
                            db_alias = db_alias_with_options.split()[0].rstrip(';')
                            try:
                                test_query = f"SELECT 1 FROM {db_alias}.information_schema.tables LIMIT 1"
                                duckdb_con.execute(test_query)
                                logger.info(f"Database '{db_alias}' is already attached, skipping ATTACH command.")
                                results[f"command_{i}"] = {"status": "skipped", "message": f"Database '{db_alias}' is already attached."}
                                continue
                            except Exception:
                                logger.info(f"Attaching database as '{db_alias}'.")
                                
                        result = duckdb_con.execute(cmd).fetchall()
                        results[f"command_{i}"] = {"status": "success", "message": f"Database attached"}
                    except Exception as attach_error:
                        logger.error(f"Error in ATTACH command: {attach_error}.")
                        results[f"command_{i}"] = {"status": "error", "message": f"ATTACH operation failed: {str(attach_error)}"}
                elif cmd.strip().upper().startswith("DETACH"):
                    try:
                        detach_parts = cmd.strip().split()
                        if len(detach_parts) >= 2:
                            detach_alias = detach_parts[1].rstrip(';')
                            logger.info(f"Detaching database '{detach_alias}'.")
                            
                        result = duckdb_con.execute(cmd).fetchall()
                        results[f"command_{i}"] = {"status": "success", "message": f"Database detached"}
                    except Exception as detach_error:
                        logger.warning(f"Error in DETACH command: {detach_error}.")
                        results[f"command_{i}"] = {"status": "warning", "message": f"DETACH operation failed: {str(detach_error)}"}
                else:
                    result = duckdb_con.execute(cmd).fetchall()
                    results[f"command_{i}"] = result

        elif bucket and blob_path and table:
            temp_table = f"temp_{table}_{int(time.time())}"
            if bucket and blob_path:
                logger.info(f"Reading data from bucket {bucket}, blob {blob_path}")
                duckdb_con.execute("INSTALL httpfs;")
                duckdb_con.execute("LOAD httpfs;")
                if bucket.startswith('gs://') or blob_path.startswith('gs://'):
                    full_path = f"{bucket}/{blob_path}" if not bucket.startswith('gs://') else f"{blob_path}"
                    if not full_path.startswith('gs://'):
                        full_path = f"gs://{full_path}"
                    logger.info(f"Reading from GCS: {full_path}")
                    duckdb_con.execute("INSTALL gcs;")
                    duckdb_con.execute("LOAD gcs;")
                elif bucket.startswith('s3://') or blob_path.startswith('s3://'):
                    full_path = f"{bucket}/{blob_path}" if not bucket.startswith('s3://') else f"{blob_path}"
                    if not full_path.startswith('s3://'):
                        full_path = f"s3://{full_path}"
                    logger.info(f"Reading from S3: {full_path}")
                    duckdb_con.execute("INSTALL s3;")
                    duckdb_con.execute("LOAD s3;")
                else:
                    full_path = file_path if file_path else os.path.join(duckdb_data_dir, blob_path)
                    logger.info(f"Reading from local file: {full_path}")

                header = task_with.get('header', False)
                header_option = "TRUE" if header else "FALSE"

                duckdb_con.execute(f"""
                    CREATE TABLE {temp_table} AS 
                    SELECT * FROM read_csv('{full_path}', header={header_option}, auto_detect=TRUE);
                """)

                row_count = duckdb_con.execute(f"SELECT COUNT(*) FROM {temp_table}").fetchone()[0]
                logger.info(f"Read {row_count} rows from {full_path} into {temp_table}")

                schema_info = duckdb_con.execute(f"DESCRIBE {temp_table}").fetchall()
                columns = []
                for col_info in schema_info:
                    col_name = col_info[0]
                    col_type = col_info[1]
                    if col_type in ('INTEGER', 'BIGINT', 'SMALLINT', 'TINYINT'):
                        pg_type = 'INTEGER'
                    elif col_type in ('DOUBLE', 'REAL', 'FLOAT'):
                        pg_type = 'DOUBLE PRECISION'
                    elif col_type == 'VARCHAR':
                        pg_type = 'TEXT'
                    elif col_type == 'BOOLEAN':
                        pg_type = 'BOOLEAN'
                    elif col_type.startswith('TIMESTAMP'):
                        pg_type = 'TIMESTAMP'
                    elif col_type.startswith('DATE'):
                        pg_type = 'DATE'
                    else:
                        pg_type = 'TEXT'
                    columns.append(f'"{col_name}" {pg_type}')

                columns_str = ', '.join(columns)
                duckdb_con.execute(f"""
                    CREATE TABLE IF NOT EXISTS postgres_db.{table} ({columns_str})
                """)
                duckdb_con.execute(f"""
                    INSERT INTO postgres_db.{table}
                    SELECT * FROM {temp_table}
                """)
                pg_row_count = duckdb_con.execute(f"SELECT COUNT(*) FROM postgres_db.{table}").fetchone()[0]
                logger.info(f"Inserted {pg_row_count} rows into PostgreSQL table {table}")
                duckdb_con.execute(f"DROP TABLE IF EXISTS {temp_table}")

                results = {
                    "source": full_path,
                    "rows_read": row_count,
                    "rows_inserted": pg_row_count,
                    "target_table": table
                }

        duckdb_con.close()
        end_time = datetime.datetime.now()
        duration = (end_time - start_time).total_seconds()

        try:
            json_results = json.dumps(results, cls=DateTimeEncoder)
            parsed_results = json.loads(json_results)

            if log_event_callback:
                log_event_callback(
                    'task_complete', task_id, task_name, 'duckdb',
                    'success', duration, context, parsed_results,
                    {'with_params': task_with}, event_id
                )

            return {
                'id': task_id,
                'status': 'success',
                'data': parsed_results
            }
        except Exception as json_error:
            logger.warning(f"Error serializing results with DateTimeEncoder: {str(json_error)}. Returning original results.")

            if log_event_callback:
                log_event_callback(
                    'task_complete', task_id, task_name, 'duckdb',
                    'success', duration, context, str(results),
                    {'with_params': task_with}, event_id
                )

            return {
                'id': task_id,
                'status': 'success',
                'data': str(results)
            }

    except Exception as e:
        if "Object of type datetime is not JSON serializable" in str(e):
            try:
                error_msg = "Original error: datetime serialization issue. Using custom encoder to handle datetime objects."
                logger.warning(error_msg)
                json_results = json.dumps(results, cls=DateTimeEncoder)
                parsed_results = json.loads(json_results)

                end_time = datetime.datetime.now()
                duration = (end_time - start_time).total_seconds()

                if log_event_callback:
                    log_event_callback(
                        'task_complete', task_id, task_name, 'duckdb',
                        'success', duration, context, parsed_results,
                        {'with_params': task_with}, event_id
                    )

                return {
                    'id': task_id,
                    'status': 'success',
                    'data': parsed_results
                }
            except Exception as json_error:
                error_msg = f"Failed to handle datetime serialization: {str(json_error)}"
                logger.error(error_msg, exc_info=True)
        else:
            error_msg = str(e)

        logger.error(f"DuckDB task execution error: {error_msg}.", exc_info=True)
        end_time = datetime.datetime.now()
        duration = (end_time - start_time).total_seconds()
        logger.debug(f"ACTION.EXECUTE_DUCKDB_TASK: Task duration={duration} seconds (error path)")

        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_DUCKDB_TASK: Writing task_error event log")
            log_event_callback(
                'task_error', task_id, task_name, 'duckdb',
                'error', duration, context, None,
                {'error': error_msg, 'with_params': task_with}, None
            )

        return {
            'id': task_id,
            'status': 'error',
            'error': error_msg
        }

def execute_postgres_task(task_config: Dict, context: Dict, jinja_env: Environment, task_with: Dict, log_event_callback=None) -> Dict:
    """
    Execute a Postgres task.

    Args:
        task_config: The task configuration
        context: The context for rendering templates
        jinja_env: The Jinja2 environment for template rendering
        task_with: The rendered 'with' parameters dictionary
        log_event_callback: A callback function to log events

    Returns:
        A dictionary of the task result
    """
    task_id = str(uuid.uuid4())
    task_name = task_config.get('task', 'postgres_task')
    start_time = datetime.datetime.now()

    try:
        commands = task_config.get('command', task_config.get('commands', []))
        if isinstance(commands, str):
            commands_rendered = render_template(jinja_env, commands, {**context, **task_with})
            commands = []
            current_command = []
            dollar_quote = False
            dollar_quote_tag = ""

            cmd_lines = []
            for line in commands_rendered.split('\n'):
                line = line.strip()
                if line and not line.startswith('--'):
                    cmd_lines.append(line)

            commands_text = ' '.join(cmd_lines)
            i = 0
            while i < len(commands_text):
                char = commands_text[i]
                if char == '$' and (i + 1 < len(commands_text)) and (commands_text[i+1].isalnum() or commands_text[i+1] == '$'):
                    j = i + 1
                    while j < len(commands_text) and (commands_text[j].isalnum() or commands_text[j] == '_' or commands_text[j] == '$'):
                        j += 1
                    optional_tag = commands_text[i:j]
                    if dollar_quote and optional_tag == dollar_quote_tag:
                        dollar_quote = False
                        dollar_quote_tag = ""
                    elif not dollar_quote:
                        dollar_quote = True
                        dollar_quote_tag = optional_tag

                    current_command.append(commands_text[i:j])
                    i = j
                    continue
                if char == ';' and not dollar_quote:
                    current_cmd = ''.join(current_command).strip()
                    if current_cmd:
                        commands.append(current_cmd)
                    current_command = []
                else:
                    current_command.append(char)
                i += 1
            current_cmd = ''.join(current_command).strip()
            if current_cmd:
                commands.append(current_cmd)

        event_id = None
        if log_event_callback:
            event_id = log_event_callback(
                'task_start', task_id, task_name, 'postgres',
                'in_progress', 0, context, None,
                {'with_params': task_with}, None
            )

        pg_host_raw = task_with.get('db_host', os.environ.get('POSTGRES_HOST', 'localhost'))
        pg_port_raw = task_with.get('db_port', os.environ.get('POSTGRES_PORT', '5434'))
        pg_user_raw = task_with.get('db_user', os.environ.get('POSTGRES_USER', 'noetl'))
        pg_password_raw = task_with.get('db_password', os.environ.get('POSTGRES_PASSWORD', 'noetl'))
        pg_db_raw = task_with.get('db_name', os.environ.get('POSTGRES_DB', 'noetl'))

        pg_host = render_template(jinja_env, pg_host_raw, context) if isinstance(pg_host_raw, str) and '{{' in pg_host_raw else pg_host_raw
        pg_port = render_template(jinja_env, pg_port_raw, context) if isinstance(pg_port_raw, str) and '{{' in pg_port_raw else pg_port_raw
        pg_user = render_template(jinja_env, pg_user_raw, context) if isinstance(pg_user_raw, str) and '{{' in pg_user_raw else pg_user_raw
        pg_password = render_template(jinja_env, pg_password_raw, context) if isinstance(pg_password_raw, str) and '{{' in pg_password_raw else pg_password_raw
        pg_db = render_template(jinja_env, pg_db_raw, context) if isinstance(pg_db_raw, str) and '{{' in pg_db_raw else pg_db_raw

        if 'db_conn_string' in task_with:
            conn_string_raw = task_with.get('db_conn_string')
            pg_conn_string = render_template(jinja_env, conn_string_raw, context) if isinstance(conn_string_raw, str) and '{{' in conn_string_raw else conn_string_raw
        else:
            pg_conn_string = f"dbname={pg_db} user={pg_user} password={pg_password} host={pg_host} port={pg_port}"

        logger.info(f"Connecting to Postgres at {pg_host}:{pg_port}/{pg_db}")

        try:
            conn = psycopg.connect(pg_conn_string)
        except Exception as e:
            safe_conn_string = pg_conn_string.replace(f"password={pg_password}", "password=***")
            logger.error(f"Failed to connect to PostgreSQL with connection string: {safe_conn_string}")
            raise

        results = {}

        if commands:
            for i, cmd in enumerate(commands):
                logger.info(f"Executing Postgres command: {cmd}")
                is_select = cmd.strip().upper().startswith("SELECT")
                is_call = cmd.strip().upper().startswith("CALL")
                returns_data = is_select or "RETURNING" in cmd.upper() or is_call

                try:
                    with conn.cursor() as cursor:
                        cursor.execute(cmd)

                        if returns_data:
                            column_names = [desc[0] for desc in cursor.description] if cursor.description else []
                            rows = cursor.fetchall()
                            result_data = []
                            for row in rows:
                                row_dict = {}
                                for i, col_name in enumerate(column_names):
                                    if isinstance(row[i], dict) or (isinstance(row[i], str) and (row[i].startswith('{') or row[i].startswith('['))):
                                        try:
                                            row_dict[col_name] = row[i]
                                        except:
                                            row_dict[col_name] = row[i]
                                    else:
                                        row_dict[col_name] = row[i]
                                result_data.append(row_dict)

                            results[f"command_{i}"] = {
                                "status": "success",
                                "rows": result_data,
                                "row_count": len(rows),
                                "columns": column_names
                            }
                        else:
                            conn.commit()
                            results[f"command_{i}"] = {
                                "status": "success",
                                "row_count": cursor.rowcount,
                                "message": f"Command executed. {cursor.rowcount} rows affected."
                            }
                except Exception as cmd_error:
                    logger.error(f"Error executing Postgres command: {cmd_error}")
                    results[f"command_{i}"] = {
                        "status": "error",
                        "message": str(cmd_error)
                    }

        conn.close()

        end_time = datetime.datetime.now()
        duration = (end_time - start_time).total_seconds()

        if log_event_callback:
            log_event_callback(
                'task_complete', task_id, task_name, 'postgres',
                'success', duration, context, results,
                {'with_params': task_with}, event_id
            )

        return {
            'id': task_id,
            'status': 'success',
            'data': results
        }

    except Exception as e:
        error_msg = str(e)
        logger.error(f"Postgres task execution error: {error_msg}", exc_info=True)
        end_time = datetime.datetime.now()
        duration = (end_time - start_time).total_seconds()

        if log_event_callback:
            log_event_callback(
                'task_error', task_id, task_name, 'postgres',
                'error', duration, context, None,
                {'error': error_msg, 'with_params': task_with}, None
            )

        return {
            'id': task_id,
            'status': 'error',
            'error': error_msg
        }

def execute_secrets_task(task_config: Dict, context: Dict, secret_manager, task_with: Dict, log_event_callback=None) -> Dict:
    """
    Execute a secret's task.

    Args:
        task_config: The task configuration
        context: The context to use for rendering templates
        secret_manager: The SecretManager instance
        task_with: The rendered 'with' parameters dictionary
        log_event_callback: A callback function to log events

    Returns:
        A dictionary of the task result
    """
    def log_event_wrapper(event_type, task_id, task_name, node_type, status, duration,
                          context, output_result, metadata, parent_event_id):
        if log_event_callback:
            if metadata is None:
                metadata = {}
            metadata['with_params'] = task_with
            return log_event_callback(
                event_type, task_id, task_name, node_type,
                status, duration, context, output_result,
                metadata, parent_event_id
            )
        return None

    return secret_manager.get_secret(task_config, context, log_event_wrapper)

def execute_task(task_config: Dict, task_name: str, context: Dict, jinja_env: Environment,
                 secret_manager=None, log_event_callback=None) -> Dict:
    """
    Execute a task type.

    Args:
        task_config: The task configuration
        task_name: The name of the task
        context: The context for rendering templates
        jinja_env: The Jinja2 environment for template rendering
        secret_manager: The SecretManager instance
        log_event_callback: A callback function to log events

    Returns:
        A dictionary of the task result
    """
    logger.debug("=== ACTION.EXECUTE_TASK: Function entry ===")
    logger.debug(f"ACTION.EXECUTE_TASK: Parameters - task_name={task_name}, task_config={task_config}")

    if not task_config:
        task_id = str(uuid.uuid4())
        error_msg = f"Task not found: {task_name}"
        logger.error(f"ACTION.EXECUTE_TASK: {error_msg}")

        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_TASK: Writing task_error event log for missing task")
            log_event_callback(
                'task_error', task_id, task_name, 'unknown',
                'error', 0, context, None,
                {'error': error_msg}, None
            )

        result = {
            'id': task_id,
            'status': 'error',
            'error': error_msg
        }
        logger.debug(f"ACTION.EXECUTE_TASK: Returning error result for missing task: {result}")
        logger.debug("=== ACTION.EXECUTE_TASK: Function exit (task not found) ===")
        return result

    task_type = task_config.get('type', 'http')
    logger.debug(f"ACTION.EXECUTE_TASK: Task type={task_type}")

    task_id = str(uuid.uuid4())
    start_time = datetime.datetime.now()
    task_with = {}

    logger.debug(f"ACTION.EXECUTE_TASK: Generated task_id={task_id}")
    logger.debug(f"ACTION.EXECUTE_TASK: Start time={start_time.isoformat()}")

    if 'with' in task_config:
        logger.debug(f"ACTION.EXECUTE_TASK: Rendering 'with' parameters: {task_config.get('with')}")
        task_with = render_template(jinja_env, task_config.get('with'), context)
        logger.debug(f"ACTION.EXECUTE_TASK: Rendered task_with: {task_with}")
        context.update(task_with)

    event_id = None
    if log_event_callback:
        logger.debug(f"ACTION.EXECUTE_TASK: Writing task_execute event log")
        event_id = log_event_callback(
            'task_execute', task_id, task_name, f'task.{task_type}',
            'in_progress', 0, context, None,
            {'task_type': task_type, 'with_params': task_with}, None
        )
        logger.debug(f"ACTION.EXECUTE_TASK: Task execute event_id={event_id}")

    logger.debug(f"ACTION.EXECUTE_TASK: Dispatching to task type handler: {task_type}")

    if task_type == 'http':
        logger.debug(f"ACTION.EXECUTE_TASK: Calling execute_http_task")
        result = execute_http_task(task_config, context, jinja_env, task_with, log_event_callback)
    elif task_type == 'python':
        logger.debug(f"ACTION.EXECUTE_TASK: Calling execute_python_task")
        result = execute_python_task(task_config, context, jinja_env, task_with, log_event_callback)
    elif task_type == 'duckdb':
        logger.debug(f"ACTION.EXECUTE_TASK: Calling execute_duckdb_task")
        result = execute_duckdb_task(task_config, context, jinja_env, task_with, log_event_callback)
    elif task_type == 'postgres':
        result = execute_postgres_task(task_config, context, jinja_env, task_with, log_event_callback)
    elif task_type == 'secrets':
        if not secret_manager:
            error_msg = "SecretManager is required for secrets tasks."

            if log_event_callback:
                log_event_callback(
                    'task_error', task_id, task_name, f'task.{task_type}',
                    'error', 0, context, None,
                    {'error': error_msg, 'with_params': task_with}, event_id
                )

            return {
                'id': task_id,
                'status': 'error',
                'error': error_msg
            }

        result = execute_secrets_task(task_config, context, secret_manager, task_with, log_event_callback)
    else:
        error_msg = f"Unsupported task type: {task_type}"
        logger.error(f"ACTION.EXECUTE_TASK: {error_msg}")

        if log_event_callback:
            logger.debug(f"ACTION.EXECUTE_TASK: Writing task_error event log for unsupported task type")
            log_event_callback(
                'task_error', task_id, task_name, f'task.{task_type}',
                'error', 0, context, None,
                {'error': error_msg, 'with_params': task_with}, event_id
            )

        result = {
            'id': task_id,
            'status': 'error',
            'error': error_msg
        }

    if 'return' in task_config and result['status'] == 'success':
        transformed_result = render_template(jinja_env, task_config['return'], {
            **context,
            'result': result['data'],
            'status': result['status']
        })

        result['data'] = transformed_result

    end_time = datetime.datetime.now()
    duration = (end_time - start_time).total_seconds()

    if log_event_callback:
        log_event_callback(
            'task_complete', task_id, task_name, f'task.{task_type}',
            result['status'], duration, context, result.get('data'),
            {'task_type': task_type, 'with_params': task_with}, event_id
        )

    return result

def report_event(event_data: Dict, server_url: str) -> Dict:
    """
    Report an event to the server.

    Args:
        event_data: The event data to report
        server_url: The URL of the server

    Returns:
        The server response
    """
    timeout = 5.0

    try:
        with httpx.Client(timeout=timeout) as client:
            response = client.post(f"{server_url}/events", json=event_data)
            response.raise_for_status()
            report_event.failure_count = 0
            return response.json()
    except Exception as e:
        if not hasattr(report_event, 'failure_count'):
            report_event.failure_count = 0
        report_event.failure_count += 1

        logger.error(f"Error reporting event to server (failure {report_event.failure_count}): {e}.")

        if report_event.failure_count > 3:
            logger.warning(f"Too many failures ({report_event.failure_count}). Disabling event reporting.")
            return {"status": "error", "error": "Event reporting disabled. Too many failures."}

        return {"status": "error", "error": str(e)}
